"""
Build hook for stripping all binary objects after building a recipe.

After the build() script is run, and before the artifacts are packaged, this
hook looks for ELF-files in the build directory and strips them. This
behavior is disabled if the recipe declares the 'nostrip' flag.
"""
import os
import logging
import shlex
from typing import Callable, List
import docker
from elftools.elf.elffile import ELFFile, ELFError
from toltec import bash
from toltec.builder import Builder
from toltec.recipe import Recipe
from toltec.util import listener

logger = logging.getLogger(__name__)

MOUNT_SRC = "/src"
TOOLCHAIN = "toolchain:v1.3.1"


def walk_elfs(src_dir: str, for_each: Callable) -> None:
    """Walk through all the ELF binaries in a directory and run a method for each of them"""
    for directory, _, files in os.walk(src_dir):
        for file_name in files:
            file_path = os.path.join(directory, file_name)

            try:
                with open(file_path, "rb") as file:
                    for_each(ELFFile(file), file_path)
            except ELFError:
                # Ignore non-ELF files
                pass
            except IsADirectoryError:
                # Ignore directories
                pass


def run_in_container(
    builder: Builder, src_dir: str, _logger: logging.Logger, script: List[str]
) -> None:
    """Run a script in a container and log output"""
    logs = bash.run_script_in_container(
        builder.docker,
        image=builder.IMAGE_PREFIX + TOOLCHAIN,
        mounts=[
            docker.types.Mount(
                type="bind",
                source=os.path.abspath(src_dir),
                target=MOUNT_SRC,
            )
        ],
        variables={},
        script="\n".join(script),
    )
    bash.pipe_logs(_logger, logs)


def register(builder: Builder) -> None:
    """Register the hook"""

    @listener(builder.post_build)
    def post_build(  # pylint: disable=too-many-locals,too-many-branches
        builder: Builder, recipe: Recipe, src_dir: str
    ) -> None:
        if "nostrip" in recipe.flags:
            logger.debug("Skipping strip ('nostrip' flag is set)")
            return

        # Search for binary objects that can be stripped
        strip_arm: List[str] = []
        strip_x86: List[str] = []

        def filter_elfs(info: ELFFile, file_path: str) -> None:
            symtab = info.get_section_by_name(".symtab")
            if not symtab:
                return
            if info.get_machine_arch() == "ARM":
                strip_arm.append(file_path)
            elif info.get_machine_arch() in ("x86", "x64"):
                strip_x86.append(file_path)

        walk_elfs(src_dir, filter_elfs)

        if not strip_arm and not strip_x86:
            logger.debug("Skipping, no binaries found")
            return

        # Save original mtimes to restore them afterwards
        # This will prevent any Makefile rules to be triggered again
        # in packaging scripts that use `make install`
        original_mtime = {}

        for file_path in strip_arm + strip_x86:
            original_mtime[file_path] = os.stat(file_path).st_mtime_ns

        # Run strip on found binaries
        script = []

        def docker_file_path(file_path: str) -> str:
            return shlex.quote(
                os.path.join(MOUNT_SRC, os.path.relpath(file_path, src_dir))
            )

        # Strip debugging symbols and unneeded sections
        if strip_x86:
            script.append(
                "strip --strip-all -- "
                + " ".join(
                    docker_file_path(file_path) for file_path in strip_x86
                )
            )

            logger.debug("x86 binaries to be stripped:")

            for file_path in strip_x86:
                logger.debug(
                    " - %s",
                    os.path.relpath(file_path, src_dir),
                )

        if strip_arm:
            script.append(
                '"${CROSS_COMPILE}strip" --strip-all -- '
                + " ".join(
                    docker_file_path(file_path) for file_path in strip_arm
                )
            )

            logger.debug("ARM binaries to be stripped:")

            for file_path in strip_arm:
                logger.debug(
                    " - %s",
                    os.path.relpath(file_path, src_dir),
                )

        run_in_container(builder, src_dir, logger, script)

        # Restore original mtimes
        for file_path, mtime in original_mtime.items():
            os.utime(file_path, ns=(mtime, mtime))



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT

import unittest
import textwrap
from datetime import datetime
from toltec.recipe import Package, Recipe
from toltec.version import (
    Version,
    Dependency,
    DependencyKind,
    VersionComparator,
)


class TestRecipe(unittest.TestCase):
    def test_derived_fields(self) -> None:
        rec = Recipe(
            path="test",
            timestamp=datetime.now(),
            sources=set(),
            makedepends=set(),
            maintainer="Test <test@example.com>",
            image="",
            arch="armv7-3.2",
            flags=[],
            prepare="",
            build="",
            packages={},
        )

        pkg = Package(
            name="test",
            parent=rec,
            version=Version(42, "12.1", "8"),
            desc="Test package",
            url="https://example.com/toltec/test",
            section="misc",
            license="MIT",
            installdepends={
                Dependency(
                    DependencyKind.HOST,
                    "test-dep",
                    VersionComparator.EQUAL,
                    Version(42, "1.0.0", "8"),
                ),
                Dependency(
                    DependencyKind.HOST,
                    "aaaaaaaa",
                    VersionComparator.GREATER_THAN_OR_EQUAL,
                    Version(0, "1.0.0", "1"),
                ),
            },
            recommends={Dependency(DependencyKind.HOST, "recommended")},
            optdepends={Dependency(DependencyKind.HOST, "optdep")},
            conflicts={Dependency(DependencyKind.HOST, "conflict")},
            replaces={Dependency(DependencyKind.HOST, "replaced")},
            provides={Dependency(DependencyKind.HOST, "provided")},
            package="",
            preinstall="",
            configure="",
            preupgrade="",
            postupgrade="",
            preremove="",
            postremove="",
        )

        self.assertEqual(pkg.pkgid(), "test_42_12.1-8_armv7-3.2")
        self.assertEqual(
            pkg.filename(), "armv7-3.2/test_42_12.1-8_armv7-3.2.ipk"
        )
        self.assertEqual(
            pkg.control_fields(),
            textwrap.dedent(
                """\
                Package: test
                Description: Test package
                Homepage: https://example.com/toltec/test
                Version: 42:12.1-8
                Section: misc
                Maintainer: Test <test@example.com>
                License: MIT
                Architecture: armv7-3.2
                Depends: aaaaaaaa (>= 1.0.0-1), test-dep (= 42:1.0.0-8)
                Recommends: recommended
                Suggests: optdep
                Conflicts: conflict
                Replaces: replaced
                Provides: provided
                """
            ),
        )



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Parse recipes from Bash files."""

import warnings
from itertools import product
from typing import Any, Dict, Generator, Iterable, Optional, Tuple
import os
import dateutil.parser
from ..version import (
    Version,
    InvalidVersionError,
    Dependency,
    InvalidDependencyError,
    DependencyKind,
)
from .. import bash
from ..recipe import (
    RecipeBundle,
    Recipe,
    Package,
    RecipeError,
    RecipeWarning,
    Source,
)


def parse(path: str) -> RecipeBundle:
    """
    Load a recipe defined as a Bash file.

    :param path: path to the directory containing the recipe definition
    :returns: loaded recipe
    """
    with open(os.path.join(path, "package"), "r", encoding="UTF-8") as recipe:
        result = {}
        definition = recipe.read()
        variables, functions = bash.get_declarations(definition)

        for arch, variables, functions in _instantiate_arch(
            path, variables, functions
        ):
            result[arch] = _parse_recipe(path, variables, functions)

        return result


def _instantiate_arch(
    path: str,
    variables: bash.Variables,
    functions: bash.Functions,
) -> Generator[Tuple[str, bash.Variables, bash.Functions], None, None]:
    """
    Instantiate a recipe definition for each supported architecture.

    :param path: path to the directory containing the recipe definition
    :param variables: Bash variables defined in the recipe
    :param functions: Bash functions defined in the recipe
    :returns: instantiated variables and functions
    """
    archs = _pop_field_indexed(path, variables, "archs", ["rmall"])
    assert archs is not None

    for arch in archs:
        loc_vars: bash.Variables = variables.copy()
        loc_funcs: bash.Functions = functions.copy()
        loc_vars["arch"] = arch

        # Merge variables suffixed with the selected architecture
        # into normal variables, drop other arch-specific declarations
        for name, value in list(loc_vars.items()):
            last_underscore = name.rfind("_")

            if last_underscore == -1:
                continue

            name_arch = name[last_underscore + 1 :]

            if name_arch not in archs:
                continue

            del loc_vars[name]

            if name_arch != arch:
                continue

            name = name[:last_underscore]

            if name not in loc_vars:
                loc_vars[name] = value
            else:
                normal_value = loc_vars[name]

                if isinstance(normal_value, str):
                    if not isinstance(value, str):
                        raise RecipeError(
                            path,
                            f"Field '{name}' was declared several times with \
different types",
                        )

                    loc_vars[name] = value

                if isinstance(normal_value, list):
                    if not isinstance(value, list):
                        raise RecipeError(
                            path,
                            f"Field '{name}' was declared several times with \
different types",
                        )

                    normal_value.extend(value)

        yield arch or "", loc_vars, loc_funcs


def _parse_recipe(  # pylint: disable=too-many-locals, disable=too-many-statements
    path: str,
    variables: bash.Variables,
    functions: bash.Functions,
) -> Recipe:
    """
    Load an architecture-specialized recipe.

    :param path: path to the directory containing the recipe definition
    :param variables: specialized Bash variables for the recipe
    :param functions: specialized Bash functions for the recipe
    :raises RecipeError: if the recipe contains an error
    :returns: loaded recipe
    """
    attrs: Dict[str, Any] = {}
    attrs["path"] = path
    raw_vars: bash.Variables = {}

    flags = raw_vars["flags"] = _pop_field_indexed(path, variables, "flags", [])
    attrs["flags"] = [flag or "" for flag in flags]

    timestamp_str = _pop_field_string(path, variables, "timestamp")
    raw_vars["timestamp"] = timestamp_str

    try:
        attrs["timestamp"] = dateutil.parser.isoparse(timestamp_str)
    except ValueError as err:
        raise RecipeError(
            path, "Field 'timestamp' does not contain a valid ISO-8601 date"
        ) from err

    sources = _pop_field_indexed(path, variables, "source", [])
    raw_vars["source"] = sources

    sha256sums = _pop_field_indexed(path, variables, "sha256sums", [])
    raw_vars["sha256sums"] = sha256sums

    noextract = _pop_field_indexed(path, variables, "noextract", [])
    raw_vars["noextract"] = noextract

    if len(sources) != len(sha256sums):
        raise RecipeError(
            path,
            f"Expected the same number of sources and checksums, got \
{len(sources)} source(s) and {len(sha256sums)} checksum(s)",
        )

    attrs["sources"] = set()

    for source, checksum in zip(sources, sha256sums):
        attrs["sources"].add(
            Source(
                url=source or "",
                checksum=checksum or "SKIP",
                noextract=os.path.basename(source or "") in noextract,
            )
        )

    makedepends_raw = _pop_field_indexed(path, variables, "makedepends", [])
    raw_vars["makedepends"] = makedepends_raw
    attrs["makedepends"] = {
        Dependency.parse(dep or "") for dep in makedepends_raw
    }

    attrs["maintainer"] = raw_vars["maintainer"] = _pop_field_string(
        path, variables, "maintainer"
    )

    attrs["image"] = raw_vars["image"] = _pop_field_string(
        path, variables, "image", ""
    )

    attrs["arch"] = raw_vars["arch"] = _pop_field_string(
        path, variables, "arch"
    )

    if attrs["image"] and "build" not in functions:
        raise RecipeError(
            path,
            "Missing build() function for a recipe which declares a \
build image",
        )

    if not attrs["image"] and "build" in functions:
        raise RecipeError(
            path,
            "Missing image declaration for a recipe which has a \
build() step",
        )

    attrs["prepare"] = functions.pop("prepare", "")
    attrs["build"] = functions.pop("build", "")

    pkgnames = _pop_field_indexed(path, variables, "pkgnames")
    attrs["packages"] = {}

    result = Recipe(**attrs)

    if len(pkgnames) == 1:
        # Single-package recipe: use global declarations
        pkg_name = pkgnames[0]
        assert pkg_name is not None
        variables["pkgname"] = pkg_name
        attrs["packages"][pkg_name] = _parse_package(
            result,
            variables.copy(),
            raw_vars.copy(),
            functions,
        )
    else:
        # Split-package recipe: load package-local declarations
        pkg_decls = {}

        for sub_pkg_name in pkgnames:
            assert sub_pkg_name is not None

            if sub_pkg_name not in functions:
                raise RecipeError(
                    path,
                    f"Missing required function {sub_pkg_name}() for \
corresponding package",
                )

            pkg_def = functions.pop(sub_pkg_name)
            context = bash.put_variables(
                {
                    **raw_vars,
                    **variables,
                    "pkgname": sub_pkg_name,
                }
            )
            pkg_decls[sub_pkg_name] = bash.get_declarations(context + pkg_def)

            for var_name in raw_vars:
                del pkg_decls[sub_pkg_name][0][var_name]

        for sub_pkg_name, (pkg_vars, pkg_funcs) in pkg_decls.items():
            attrs["packages"][sub_pkg_name] = _parse_package(
                result,
                pkg_vars,
                raw_vars.copy(),
                {**functions, **pkg_funcs},
            )

    _add_script_header(
        result,
        ("prepare", "build"),
        {**raw_vars, **variables},
        functions,
    )

    return result


def _parse_package(  # pylint: disable=too-many-locals, disable=too-many-statements
    parent: Recipe,
    variables: bash.Variables,
    raw_vars: bash.Variables,
    functions: bash.Functions,
) -> Package:
    """
    Load a package.

    :param parent: specialized recipe which declares this package
    :param variables: variables declared in the package
    :param raw_vars: variables used to populate fields in the parent recipe
    :param functions: functions declared in the package
    :raises RecipeError: if the package contains an error
    """
    attrs: Dict[str, Any] = {}
    attrs["parent"] = parent

    # Parse fields
    attrs["name"] = raw_vars["pkgname"] = _pop_field_string(
        parent.path, variables, "pkgname"
    )

    pkgver_str = _pop_field_string(parent.path, variables, "pkgver")
    raw_vars["pkgver"] = pkgver_str

    try:
        attrs["version"] = Version.parse(pkgver_str)
    except InvalidVersionError as err:
        raise RecipeError(
            parent.path, f"Failed to parse version number: '{pkgver_str}'"
        ) from err

    attrs["desc"] = raw_vars["pkgdesc"] = _pop_field_string(
        parent.path, variables, "pkgdesc"
    )

    attrs["url"] = raw_vars["url"] = _pop_field_string(
        parent.path, variables, "url"
    )

    attrs["section"] = raw_vars["section"] = _pop_field_string(
        parent.path, variables, "section"
    )

    attrs["license"] = raw_vars["license"] = _pop_field_string(
        parent.path, variables, "license"
    )

    for field in (
        "installdepends",
        "recommends",
        "optdepends",
        "conflicts",
        "replaces",
        "provides",
    ):
        field_raw = _pop_field_indexed(parent.path, variables, field, [])
        raw_vars[field] = field_raw
        attrs[field] = set()

        for dep_raw in field_raw:
            assert dep_raw is not None
            try:
                dep = Dependency.parse(dep_raw)
            except (InvalidVersionError, InvalidDependencyError) as err:
                raise RecipeError(
                    parent.path, f"Failed to parse dependency: '{dep_raw}'"
                ) from err

            if dep.kind != DependencyKind.HOST:
                raise RecipeError(
                    parent.path,
                    f"Only host packages are supported in the \
'{field}' field, cannot add dependency '{dep}'",
                )

            attrs[field].add(dep)

    # Parse functions
    attrs["package"] = functions.pop("package")

    for action in ("preinstall", "configure"):
        attrs[action] = functions.pop(action, "")

    for rel, step in product(("pre", "post"), ("remove", "upgrade")):
        attrs[rel + step] = functions.pop(rel + step, "")

    # Check that remaining variables and functions are prefixed
    for var_name in variables.keys():
        if not var_name.startswith("_"):
            warnings.warn(
                f"{parent.path}: Unknown field '{var_name}'. Make sure to \
prefix the names of custom fields with '_'",
                RecipeWarning,
            )

    for func_name in functions.keys():
        if not func_name.startswith("_"):
            warnings.warn(
                f"{parent.path}: Unknown function '{func_name}'. Make sure to \
prefix the names of custom functions with '_'",
                RecipeWarning,
            )

    result = Package(**attrs)
    _add_script_header(
        result,
        (
            "package",
            "preinstall",
            "configure",
            "preremove",
            "postremove",
            "preupgrade",
            "postupgrade",
        ),
        {**raw_vars, **variables},
        functions,
    )
    return result


def _pop_field_string(
    path: str,
    variables: bash.Variables,
    name: str,
    default: Optional[str] = None,
) -> str:
    if name not in variables:
        if default is None:
            raise RecipeError(path, f"Missing required field '{name}'")
        return default

    value = variables.pop(name)

    if not isinstance(value, str):
        raise RecipeError(
            path,
            f"Field '{name}' must be a string, got a {type(value).__name__}",
        )

    return value


def _pop_field_indexed(
    path: str,
    variables: bash.Variables,
    name: str,
    default: Optional[bash.IndexedArray] = None,
) -> bash.IndexedArray:
    if name not in variables:
        if default is None:
            raise RecipeError(path, f"Missing required field '{name}'")
        return default

    value = variables.pop(name)

    if not isinstance(value, list):
        _name = type(value).__name__
        raise RecipeError(
            path,
            f"Field '{name}' must be an indexed array, got a {_name}",
        )

    return value


def _add_script_header(
    obj: object,
    keys: Iterable[str],
    variables: bash.Variables,
    functions: bash.Functions,
) -> None:
    header = "\n".join(
        (
            bash.put_variables(variables),
            bash.put_functions(functions),
        )
    )

    for key in keys:
        if getattr(obj, key):
            setattr(obj, key, header + getattr(obj, key))



# Copyright (c) 2023 The Toltec Contributors
# SPDX-License-Identifier: MIT

import unittest
import subprocess
import shutil

from os import path
from tempfile import TemporaryDirectory
from toltec.hooks.strip import walk_elfs
from elftools.elf.elffile import ELFFile


class TestBuild(unittest.TestCase):
    def setUp(self) -> None:
        self.dir = path.dirname(path.realpath(__file__))
        self.fixtures_dir = path.join(self.dir, "fixtures")

    def test_strip(self) -> None:
        with TemporaryDirectory() as tmp_dir:
            rec_dir = path.join(self.fixtures_dir, "hello")
            work_dir = path.join(tmp_dir, "build")
            dist_dir = path.join(tmp_dir, "dist")

            result = subprocess.run(
                [
                    "python3",
                    "-m",
                    "toltec",
                    "--work-dir",
                    work_dir,
                    "--dist-dir",
                    dist_dir,
                    "--",
                    rec_dir,
                ],
                capture_output=True,
                check=False,
            )
            self.assertEqual(
                result.returncode, 0, result.stderr.decode("utf-8")
            )
            self.assertEqual(result.stdout.decode("utf-8"), "")
            walk_elfs(
                work_dir,
                lambda i, p: self.assertIsNone(
                    i.get_section_by_name(".symtab"), f"{p} is not stripped"
                ),
            )

        with TemporaryDirectory() as tmp_dir:
            src_dir = path.join(self.fixtures_dir, "hello")
            rec_dir = path.join(tmp_dir, "src")
            work_dir = path.join(tmp_dir, "build")
            dist_dir = path.join(tmp_dir, "dist")
            shutil.copytree(src_dir, rec_dir)
            replacements = {"flags=()": "flags=(nostrip)"}
            with open(
                path.join(src_dir, "package"), "rt", encoding="utf-8"
            ) as infile, open(
                path.join(rec_dir, "package"), "wt", encoding="utf-8"
            ) as outfile:
                for line in infile:
                    for src, target in replacements.items():
                        line = line.replace(src, target)
                    outfile.write(line)

            result = subprocess.run(
                [
                    "python3",
                    "-m",
                    "toltec",
                    "--work-dir",
                    work_dir,
                    "--dist-dir",
                    dist_dir,
                    "--",
                    rec_dir,
                ],
                capture_output=True,
                check=False,
            )
            self.assertEqual(
                result.returncode, 0, result.stderr.decode("utf-8")
            )
            self.assertEqual(result.stdout.decode("utf-8"), "")
            walk_elfs(
                work_dir,
                lambda i, p: self.assertIsNotNone(
                    i.get_section_by_name(".symtab"), f"{p} is stripped"
                ),
            )



"""
Build hook for automatically reloading applications in oxide

After the artifacts are packaged, this hook looks for files in either
/opt/etc/draft or /opt/usr/share/applications and adds reload-oxide-apps to
configure, postupgrade, and postremove
"""
import os
import logging

from toltec.builder import Builder
from toltec.recipe import Package
from toltec.util import listener

logger = logging.getLogger(__name__)

OXIDE_HOOK = "\nreload-oxide-apps\n"


def register(builder: Builder) -> None:
    """Register the hook"""

    @listener(builder.post_package)
    def post_package(
        builder: Builder,  # pylint: disable=unused-argument
        package: Package,
        src_dir: str,  # pylint: disable=unused-argument
        pkg_dir: str,
    ) -> None:
        if os.path.exists(
            os.path.join(pkg_dir, "opt/usr/share/applications")
        ) or os.path.exists(os.path.join(pkg_dir, "opt/etc/draft")):
            package.configure += OXIDE_HOOK
            package.postupgrade += OXIDE_HOOK
            package.postremove += OXIDE_HOOK



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""
Repository management.
"""

import gzip
import locale
import logging
import os
import textwrap
from .util import file_sha256
from . import ipk

logger = logging.getLogger(__name__)


def make_index(base_dir: str, _start: bool = True) -> None:
    """
    Recursively generate index files for all the packages in folder.

    Traverses :param:`base_dir` and all its subfolders and creates a package
    index for each subfolder containing the metadata of ipk packages
    contained directly in that folder.

    :param base_dir: directory to start the traversal from
    """
    if _start:
        logger.info("Generating package index")

    index_path = os.path.join(base_dir, "Packages")
    index_gzip_path = os.path.join(base_dir, "Packages.gz")

    with open(
        index_path, "w", encoding=locale.getencoding()
    ) as index_file, gzip.open(index_gzip_path, "wt") as index_gzip_file:
        for entry in os.scandir(base_dir):
            if entry.name in ("Packages", "Packages.gz"):
                pass
            elif entry.is_dir():
                make_index(entry.path, _start=False)
            elif entry.is_file() and entry.name.endswith(".ipk"):
                with ipk.Reader(entry.path) as package:
                    metadata = package.metadata
                    assert metadata is not None

                metadata += textwrap.dedent(
                    f"""\
                    Filename: {entry.name}
                    SHA256sum: {file_sha256(entry.path)}
                    Size: {os.path.getsize(entry.path)}

                    """
                )

                index_file.write(metadata)
                index_gzip_file.write(metadata)



# Copyright (c) 2022 The Toltec Contributors
# SPDX-License-Identifier: MIT

import unittest
import tarfile
from tempfile import TemporaryDirectory
import os
from toltec import ipk
from io import BytesIO


test_package = {
    "epoch": 1337,
    "metadata": "Test: OK\n",
    "scripts": {"postinst": "echo OK"},
    "data": [
        {
            "type": "file",
            "path": "./file1",
            "contents": "file 1\ntest\n",
            "mode": 0o644,
        },
        {
            "type": "dir",
            "path": "./zsubfolder",
            "mode": 0o755,
        },
        {
            "type": "file",
            "path": "./zsubfolder/file2",
            "contents": "this is\nfile 2",
            "mode": 0o400,
        },
    ],
}


class TestIpk(unittest.TestCase):
    def setUp(self):
        # Create data tree
        self.temp_dir = TemporaryDirectory()
        temp = self.temp_dir.name

        for member in test_package["data"]:
            if member["type"] == "file":
                with open(temp + "/" + member["path"], "w") as file:
                    file.write(member["contents"])

                os.chmod(temp + "/" + member["path"], member["mode"])
            else:
                os.mkdir(temp + "/" + member["path"], member["mode"])

    def tearDown(self):
        self.temp_dir.cleanup()

    def test_write_control(self):
        output = BytesIO()
        ipk.write_control(
            output,
            test_package["epoch"],
            test_package["metadata"],
            test_package["scripts"],
        )
        output.seek(0)

        with tarfile.TarFile.open(fileobj=output) as archive:
            members = archive.getmembers()
            self.assertEqual(len(members), 3)

            self.assertTrue(members[0].isdir())
            self.assertEqual(members[0].name, ".")
            self.assertEqual(members[0].size, 0)
            self.assertEqual(members[0].mode, 0o755)

            self.assertTrue(members[1].isfile())
            self.assertEqual(members[1].name, "./control")
            self.assertEqual(members[1].size, len(test_package["metadata"]))
            self.assertEqual(members[1].mode, 0o644)
            self.assertEqual(
                archive.extractfile(members[1]).read().decode(),
                test_package["metadata"],
            )

            self.assertTrue(members[2].isfile())
            self.assertEqual(members[2].name, "./postinst")
            self.assertEqual(
                members[2].size, len(test_package["scripts"]["postinst"])
            )
            self.assertEqual(members[2].mode, 0o755)
            self.assertEqual(
                archive.extractfile(members[2]).read().decode(),
                test_package["scripts"]["postinst"],
            )

            for member in members:
                self.assertEqual(member.uid, 0)
                self.assertEqual(member.gid, 0)
                self.assertEqual(member.uname, "")
                self.assertEqual(member.gname, "")
                self.assertEqual(member.mtime, test_package["epoch"])

    def test_write_data_empty(self):
        output = BytesIO()
        ipk.write_data(output, test_package["epoch"])
        output.seek(0)

        with tarfile.TarFile.open(fileobj=output) as archive:
            self.assertEqual(archive.getmembers(), [])

    def test_write_data_dir(self):
        output = BytesIO()
        ipk.write_data(output, test_package["epoch"], self.temp_dir.name)
        output.seek(0)

        with tarfile.TarFile.open(fileobj=output) as archive:
            members = archive.getmembers()
            self.assertEqual(len(members), 4)

            self.assertTrue(members[0].isdir())
            self.assertEqual(members[0].name, ".")
            self.assertEqual(members[0].size, 0)
            self.assertEqual(members[0].mode, 0o700)

            for expect, member in zip(test_package["data"], members[1:]):
                if expect["type"] == "file":
                    self.assertTrue(member.isfile())
                    self.assertEqual(member.name, expect["path"])
                    self.assertEqual(member.size, len(expect["contents"]))
                    self.assertEqual(member.mode, expect["mode"])
                    self.assertEqual(
                        archive.extractfile(member).read().decode(),
                        expect["contents"],
                    )
                else:
                    self.assertTrue(member.isdir())
                    self.assertEqual(member.name, expect["path"])
                    self.assertEqual(member.size, 0)
                    self.assertEqual(member.mode, expect["mode"])

                self.assertEqual(member.uid, 0)
                self.assertEqual(member.gid, 0)
                self.assertEqual(member.uname, "")
                self.assertEqual(member.gname, "")
                self.assertEqual(member.mtime, test_package["epoch"])

    def test_write_read(self):
        output = BytesIO()
        ipk.write(
            output,
            test_package["epoch"],
            test_package["metadata"],
            test_package["scripts"],
            self.temp_dir.name,
        )

        output.seek(0)

        with ipk.Reader(output) as result:
            self.assertEqual(result.metadata, test_package["metadata"])
            self.assertEqual(result.scripts, test_package["scripts"])
            self.assertEqual(
                result.data.getnames(),
                ["."] + [member["path"] for member in test_package["data"]],
            )



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""
Parse versions and dependency specifications.

Syntax and comparison rules for version numbers and comparison operators are
taken from Debian’s. See:

* <https://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-version>
* <https://www.debian.org/doc/debian-policy/ch-relationships.html>
"""

import re
from functools import total_ordering
from enum import Enum
from typing import Optional, Callable

# Characters permitted in the upstream part of a version number
_UPSTREAM_CHARS = "A-Za-z0-9.+~-"
_UPSTREAM_REGEX = re.compile(f"^[{_UPSTREAM_CHARS}]*$")

# Characters permitted in the revision part of a version number
_REVISION_CHARS = "A-Za-z0-9.+~"
_REVISION_REGEX = re.compile(f"^[{_REVISION_CHARS}]*$")

# Characters making up a version comparator
_COMPARATOR_CHARS = re.compile("[<>=]+")

# Regex used to find digit and non-digit chars
_DIGIT_REGEX = re.compile("[0-9]")
_NON_DIGIT_REGEX = re.compile("[^0-9]")

# Sorting key used for non-digit version parts
_ALPHA_SORT_KEY = (
    (
        "~",  # ~ sorts lower than anything, even empty parts
        None,
    )
    + tuple(chr(letter) for letter in range(ord("A"), ord("Z") + 1))
    + tuple(chr(letter) for letter in range(ord("a"), ord("z") + 1))
    + (
        "+",
        "-",
        ".",
    )
)


def _find_digit(string: str) -> int:
    """
    Find the index of the first digit char in a string, or return the
    length of the string if there is none.
    """
    matches = _DIGIT_REGEX.search(string)
    return len(string) if not matches else matches.start()


def _find_non_digit(string: str) -> int:
    """
    Find the index of the first non-digit char in a string, or return the
    length of the string if there is none.
    """
    matches = _NON_DIGIT_REGEX.search(string)
    return len(string) if not matches else matches.start()


def _compare_version_parts(left: str, right: str) -> bool:
    """
    Compare two parts of a version string according to Debian version
    sorting rules.
    :returns: true if and only if `left` is strictly lower than `right`
    """

    def split(
        string: str, index_finder: Callable[[str], int]
    ) -> tuple[str, str]:
        index = index_finder(string)
        return string[:index], string[index:]

    def map_alpha(string: str, length: int) -> tuple[int, ...]:
        return tuple(
            _ALPHA_SORT_KEY.index(string[i])
            if i in range(len(string))
            else _ALPHA_SORT_KEY.index(None)
            for i in range(length)
        )

    # Split the strings into alternating non-digit parts and numeric parts,
    # compare non-digit parts using _ALPHA_SORT_KEY and numeric parts as
    # integers, stop at the first non-equal part
    while left or right:
        left_alpha, left = split(left, _find_digit)
        right_alpha, right = split(right, _find_digit)

        max_len = max(len(left_alpha), len(right_alpha))
        left_map = map_alpha(left_alpha, max_len)
        right_map = map_alpha(right_alpha, max_len)

        if left_map != right_map:
            return left_map < right_map

        left_digit, left = split(left, _find_non_digit)
        right_digit, right = split(right, _find_non_digit)

        left_numeric = int(left_digit)
        right_numeric = int(right_digit)

        if left_numeric != right_numeric:
            return left_numeric < right_numeric

    return False


class VersionComparator(Enum):
    """Operators used to compare two version numbers."""

    LOWER_THAN = "<<"
    LOWER_THAN_OR_EQUAL = "<="
    EQUAL = "="
    GREATER_THAN_OR_EQUAL = ">="
    GREATER_THAN = ">>"


class InvalidVersionError(Exception):
    """Raised when parsing of an invalid version is attempted."""


@total_ordering
class Version:
    """
    Parse package versions.

    for details about the format and the comparison rules.
    """

    def __init__(self, epoch: int, upstream: str, revision: str):
        self.upstream = upstream
        self.revision = revision
        self.epoch = epoch

        if epoch < 0:
            raise InvalidVersionError(
                f"Invalid epoch '{epoch}', only non-negative values "
                "are allowed"
            )

        if not upstream:
            raise InvalidVersionError("Upstream version cannot be empty")

        if _UPSTREAM_REGEX.fullmatch(upstream) is None:
            raise InvalidVersionError(
                f"Invalid chars in upstream version '{upstream}', allowed "
                f"chars are {_UPSTREAM_CHARS}"
            )

        if not revision:
            raise InvalidVersionError("Revision cannot be empty")

        if _REVISION_REGEX.fullmatch(revision) is None:
            raise InvalidVersionError(
                f"Invalid chars in revision '{revision}' allowed chars "
                f"are {_REVISION_CHARS}"
            )

        self._original: Optional[str] = None

    @staticmethod
    def parse(version: str) -> "Version":
        """Parse a version number."""
        original = version
        first_colon = version.find(":")

        if first_colon == -1:
            epoch = 0
        else:
            epoch_text = version[:first_colon]
            version = version[first_colon + 1 :]

            try:
                epoch = int(epoch_text)
            except ValueError as err:
                raise InvalidVersionError(
                    f"Invalid epoch '{epoch_text}', must be numeric"
                ) from err

        last_dash = version.rfind("-")

        if last_dash == -1:
            revision = "0"
        else:
            revision = version[last_dash + 1 :]
            version = version[:last_dash]

        upstream = version

        result = Version(epoch, upstream, revision)
        result._original = original  # pylint:disable=protected-access
        return result

    def __str__(self) -> str:
        if self._original is not None:
            # Use the original parsed version string
            return self._original

        epoch = "" if self.epoch == 0 else f"{self.epoch}:"
        revision = (
            ""
            if self.revision == "0" and "-" not in self.upstream
            else f"-{self.revision}"
        )

        return f"{epoch}{self.upstream}{revision}"

    def __repr__(self) -> str:
        return f"Version(upstream={repr(self.upstream)}, \
revision={repr(self.revision)}, epoch={repr(self.epoch)})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        return (
            self.epoch == other.epoch
            and self.upstream == other.upstream
            and self.revision == other.revision
        )

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        if self.epoch != other.epoch:
            return self.epoch < other.epoch

        if self.upstream != other.upstream:
            return _compare_version_parts(self.upstream, other.upstream)

        return _compare_version_parts(self.revision, other.revision)

    def __hash__(self) -> int:
        return hash((self.epoch, self.upstream, self.revision))


class DependencyKind(Enum):
    """Kinds of dependencies that may be requested by a package."""

    # Dependency installed in the system used to build a package
    # (e.g., a Debian package)
    BUILD = "build"
    # Dependency installed alongside a package
    # (e.g., another Entware or Toltec package)
    HOST = "host"


class InvalidDependencyError(Exception):
    """Raised when parsing an invalid dependency specification."""


class Dependency:
    """
    Parse version-constrained dependencies.

    Toltec dependencies are declared using the following format:

        [host:|build:]package[(<<|<=|=|=>|>>)version]

    Dependencies of a package that start with `build:` correspond to packages
    that must be installed in the build system. Dependencies that start with
    `host:` or do not have a prefix correspond to packages that must be
    installed alongside the built package, either in the host sysroot when
    building the package, or in the target device when using it.
    """

    def __init__(
        self,
        kind: DependencyKind,
        package: str,
        version_comparator: VersionComparator = VersionComparator.EQUAL,
        version: Optional[Version] = None,
    ):
        self.kind = kind
        self.package = package
        self.version_comparator = version_comparator
        self.version = version

        self._original: Optional[str] = None

    @staticmethod
    def parse(dependency: str) -> "Dependency":
        """Parse a dependency specification."""
        original = dependency
        comp_match = _COMPARATOR_CHARS.search(dependency)

        if comp_match is None:
            version_comparator = VersionComparator.EQUAL
            version = None
        else:
            comparator_str = comp_match.group(0)
            comp_char = comp_match.start()

            for enum_comparator in VersionComparator:
                if enum_comparator.value == comparator_str:
                    version_comparator = enum_comparator
                    version = Version.parse(dependency[comp_match.end() :])
                    dependency = dependency[: comp_match.start()]
                    break
            else:
                raise InvalidDependencyError(
                    f"Invalid version comparator '{comp_char}', valid types "
                    "are "
                    + ",".join(
                        f"'{enum_comparator.value}'"
                        for enum_comparator in VersionComparator
                    )
                )

        colon = dependency.find(":")

        if colon == -1:
            kind = DependencyKind.HOST
            package = dependency
        else:
            dep_kind_str = dependency[:colon]

            for enum_kind in DependencyKind:
                if enum_kind.value == dep_kind_str:
                    kind = enum_kind
                    package = dependency[colon + 1 :]
                    break
            else:
                raise InvalidDependencyError(
                    f"Unknown dependency type '{dep_kind_str}', valid types "
                    "are "
                    + ",".join(
                        f"'{enum_kind.value}'" for enum_kind in DependencyKind
                    )
                )

        result = Dependency(kind, package, version_comparator, version)
        result._original = original  # pylint:disable=protected-access
        return result

    def match(self, version: Version) -> bool:
        """Check whether a given version fulfills this dependency."""
        if self.version is None:
            return True

        if self.version_comparator == VersionComparator.EQUAL:
            return version == self.version

        if self.version_comparator == VersionComparator.LOWER_THAN:
            return version < self.version

        if self.version_comparator == VersionComparator.LOWER_THAN_OR_EQUAL:
            return version <= self.version

        if self.version_comparator == VersionComparator.GREATER_THAN:
            return version > self.version

        # self.version_comparator == VersionComparator.GREATER_THAN_OR_EQUAL:
        return version >= self.version

    def to_debian(self) -> str:
        """Convert a dependency specification to the Debian format."""
        if self.version is None:
            return self.package

        return f"{self.package} ({self.version_comparator.value} \
{self.version})"

    def __str__(self) -> str:
        if self._original is not None:
            # Use the original parsed dependency specification
            return self._original

        kind = "build:" if self.kind == DependencyKind.BUILD else "host:"

        if self.version is None:
            return f"{kind}{self.package}"

        return f"{kind}{self.package}{self.version_comparator.value}\
{self.version}"

    def __repr__(self) -> str:
        return f"Dependency(kind={repr(self.kind)}, \
package={repr(self.package)}, \
version_comparator={repr(self.version_comparator)}, \
version={repr(self.version)})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Dependency):
            return (
                self.kind == other.kind
                and self.package == other.package
                and self.version_comparator == other.version_comparator
                and self.version == other.version
            )

        return False

    def __hash__(self) -> int:
        return hash(
            (self.kind, self.package, self.version_comparator, self.version)
        )



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Collection of useful functions."""

import argparse
from collections.abc import Iterable
import hashlib
import logging
import itertools
import functools
import os
import shutil
import sys
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    IO,
    List,
    Optional,
    Protocol,
    Type,
    Union,
)
import warnings
import zipfile
import tarfile

# Date format used in HTTP headers such as Last-Modified
HTTP_DATE_FORMAT = "%a, %d %b %Y %H:%M:%S %Z"

# Logging format for build scripts
LOGGING_FORMAT = "[%(levelname)8s] %(name)s: %(message)s"


def argparse_add_verbose(parser: argparse.ArgumentParser) -> None:
    """Add a CLI option for setting the verbosity level."""
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.DEBUG,
        default=logging.INFO,
        help="show debugging information",
    )


def argparse_add_warning(parser: argparse.ArgumentParser) -> None:
    """Add a CLI option for controlling warnings."""
    parser.add_argument(
        "-W",
        "--warnings",
        choices=("default", "error", "ignore"),
        default="default",
        help="""control warnings. 'default' sends warnings to stderr and
removes duplicate warnings, 'ignore' silences any warning, and 'error' turns
warnings into exceptions""",
    )


def setup_logging(args: argparse.Namespace) -> None:
    """Configure logging and warning control based on passed CLI flags."""
    if hasattr(args, "verbose"):
        logging.basicConfig(format=LOGGING_FORMAT, level=args.verbose)

    def formatwarning(
        message: Union[str, Warning],
        category: Type[Warning],
        filename: str,
        lineno: int,
        line: Optional[str] = None,
    ) -> str:
        del filename, lineno, line
        return f"[{category.__name__}] {message}"

    warnings.formatwarning = formatwarning
    logging.captureWarnings(True)

    if hasattr(args, "warnings"):
        warnings.simplefilter(args.warnings)


def file_sha256(path: str) -> str:
    """Compute the SHA-256 checksum of a file."""
    sha256 = hashlib.sha256()
    buffer = bytearray(128 * 1024)
    view = memoryview(buffer)

    with open(path, "rb", buffering=0) as file:
        for length in iter(lambda: file.readinto(view), 0):  # type:ignore
            sha256.update(view[:length])

    return sha256.hexdigest()


def split_all_parts(path: str) -> List[str]:
    """Split a file path into all its directory components."""
    parts = []
    prefix = path

    while prefix not in ("", "/"):
        prefix, base = os.path.split(prefix)
        if base:
            parts.append(base)

    parts.reverse()
    return parts


def split_all_exts(path: str) -> List[str]:
    """Get the list of extensions in a file path."""
    exts = []
    remaining = path

    while True:
        remaining, ext = os.path.splitext(remaining)
        if ext:
            exts.append(ext)
        else:
            break

    return exts


def all_equal(seq: Iterable) -> bool:
    """Check that all elements of a sequence are equal."""
    grouped = itertools.groupby(seq)
    first = next(grouped, (None, grouped))
    second = next(grouped, None)
    return first and not second


def remove_prefix(filenames: List[str]) -> Dict[str, str]:
    """Find and remove the longest directory prefix shared by all files."""
    split_filenames = [split_all_parts(filename) for filename in filenames]

    # Find the longest directory prefix shared by all files
    min_len = min(len(filename) for filename in split_filenames)
    prefix = 0

    while prefix < min_len and all_equal(
        filename[prefix] for filename in split_filenames
    ):
        prefix += 1

    # If there’s only one file, keep the last component
    if len(filenames) == 1:
        prefix -= 1

    mapping = {}

    for filename, split_filename in zip(filenames, split_filenames):
        if split_filename[prefix:]:
            mapping[filename] = os.path.join(*split_filename[prefix:])

    return mapping


def auto_extract(archive_path: str, dest_path: str) -> bool:
    """
    Automatically extract an archive and strip useless components.

    :param archive_path: path to the archive to extract
    :param dest_path: destination folder for the archive contents
    :returns: true if something was extracted, false if not a supported archive
    """
    exts = split_all_exts(archive_path)

    if not exts:
        return False

    if exts[0] == ".zip":
        with zipfile.ZipFile(archive_path) as zip_archive:
            _auto_extract(
                zip_archive.namelist(),
                zip_archive.getinfo,
                zip_archive.open,
                lambda member: member.is_dir(),
                lambda member: False,
                lambda member: member.external_attr >> 16 & 0x1FF,
                dest_path,
            )
        return True

    if exts[0] == ".tar" or (
        len(exts) >= 2
        and exts[0] in (".gz", ".bz2", ".xz")
        and exts[1] == ".tar"
    ):
        with tarfile.open(archive_path, mode="r") as tar_archive:
            _auto_extract(
                tar_archive.getnames(),
                tar_archive.getmember,
                tar_archive.extractfile,
                lambda member: member.isdir(),
                lambda member: member.issym(),
                lambda member: member.mode,
                dest_path,
            )
        return True

    return False


def _auto_extract(  # pylint:disable=too-many-arguments,disable=too-many-locals
    members: List[str],
    getinfo: Callable[[str], Any],
    extract: Callable[[Any], Optional[IO[bytes]]],
    isdir: Callable[[Any], bool],
    issym: Callable[[Any], bool],
    getmode: Callable[[Any], int],
    dest_path: str,
) -> None:
    """
    Generic implementation of automatic archive extraction.

    :param members: list of members of the archive
    :param getinfo: get an entry object from an entry name in the archive
    :param extract: get a reading stream corresponding to an archive entry
    :param isdir: get whether an entry is a directory or not
    :param issym: get whether an entry is a symbolic link or not
    :param getmode: get the permission bits for an entry
    :param destpath: destinatio folder for the archive contents
    """
    stripped_map = remove_prefix(members)

    for filename, stripped in stripped_map.items():
        member = getinfo(filename)
        file_path = os.path.join(dest_path, stripped)

        if isdir(member):
            os.makedirs(file_path, exist_ok=True)
        else:
            if issym(member):
                os.symlink(member.linkname, file_path)
            else:
                basedir = os.path.dirname(file_path)
                if not os.path.exists(basedir):
                    os.makedirs(basedir, exist_ok=True)

                source = extract(member)
                assert source is not None

                with source, open(file_path, "wb") as target:
                    shutil.copyfileobj(source, target)

                mode = getmode(member)
                if mode != 0:
                    os.chmod(file_path, mode)


def query_user(
    question: str,
    default: str,
    options: Optional[List[str]] = None,
    aliases: Optional[Dict[str, str]] = None,
) -> str:
    """
    Ask the user to make a choice.

    :param question: message to display before the choice
    :param default: default choice if the user inputs an empty string
    :param options: list of valid options (should be lowercase strings)
    :param aliases: accepted aliases for the valid options
    :returns: option chosen by the user
    """
    options = options or ["y", "n"]
    aliases = aliases or {"yes": "y", "no": "n"}

    if default not in options:
        raise ValueError(f"Default value {default} is not a valid option")

    prompt = "/".join(
        option if option != default else option.upper() for option in options
    )

    while True:
        sys.stdout.write(f"{question} [{prompt}] ")
        choice = input().lower()

        if not choice:
            return default

        if choice in options:
            return choice

        if choice in aliases:
            return aliases[choice]

        print("Invalid answer. Please choose among the valid options.")


def check_directory(path: str, message: str) -> bool:
    """
    Create a directory and ask the user what to do if it already exists.

    :param path: path to the directory to create
    :param message: message to display before asking the user interactively
    :returns: false if the user chose to cancel the current operation
    """
    try:
        os.mkdir(path)
    except FileExistsError:
        ans = query_user(
            message,
            default="c",
            options=["c", "r", "k"],
            aliases={
                "cancel": "c",
                "remove": "r",
                "keep": "k",
            },
        )

        if ans == "c":
            return False

        if ans == "r":
            shutil.rmtree(path)
            os.mkdir(path)

    return True


def list_tree(root: str) -> List[str]:
    """
    Get a sorted list of all files and folders under a given root folder.

    :param root: root folder to start from
    :returns: sorted list of items under the root folder
    """
    result = []

    for directory, _, files in os.walk(root):
        result.append(directory)
        for file in files:
            result.append(os.path.join(directory, file))

    return sorted(result)


HookTrigger = Callable[..., None]
HookListener = Callable[..., None]


class Hook(Protocol):  # pylint:disable=too-few-public-methods
    """Protocol for hooks."""

    @staticmethod
    def register(new_listener: HookListener) -> None:
        """Add a new listener to this hook."""

    # Invoke all listeners for this hook
    __call__: HookTrigger


def hook(func: HookTrigger) -> Hook:
    """
    Decorator for turning a function into a hook.

    :param func: empty function to declare as a hook
    :returns: usable hook
    """
    listeners: List[HookListener] = []

    def register(new_listener: HookListener) -> None:
        listeners.append(new_listener)

    @functools.wraps(func)
    def call(*args: Any, **kwargs: Any) -> None:
        for call_listener in listeners:
            call_listener(*args, **kwargs)

    setattr(call, "register", register)
    return cast(Hook, call)


def listener(target_hook: Hook) -> Callable[[HookListener], HookListener]:
    """
    Decorator for subscribing a function to a hook.

    :param target_hook: hook to add this function as a listener of
    :returns: identity
    """

    def decorator(new_listener: HookListener) -> HookListener:
        target_hook.register(new_listener)
        return new_listener

    return decorator



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Build recipes and create packages."""

import shutil
from typing import List, Mapping, Optional, Type
from types import TracebackType
import re
import os
import logging
import textwrap
from importlib.util import find_spec, module_from_spec
import docker
import requests
from . import hooks
from . import bash, util, ipk
from .recipe import RecipeBundle, Recipe, Package
from .version import DependencyKind

logger = logging.getLogger(__name__)


class BuildError(Exception):
    """Raised when a build step fails."""


class Builder:  # pylint: disable=too-few-public-methods
    """Helper class for building recipes."""

    # Detect non-local paths
    URL_REGEX = re.compile(r"[a-z]+://")

    # Prefix for all Toltec Docker images
    IMAGE_PREFIX = "ghcr.io/toltec-dev/"

    def __init__(self, work_dir: str, dist_dir: str) -> None:
        """
        Create a builder helper.

        :param work_dir: directory where packages are built
        :param dist_dir: directory where built packages are stored
        """
        self.work_dir = work_dir
        self.dist_dir = dist_dir

        try:
            self.docker = docker.from_env()
        except docker.errors.DockerException as err:
            raise BuildError(
                "Unable to connect to the Docker daemon. \
Please check that the service is running and that you have the necessary \
permissions."
            ) from err

        for hook in hooks.__all__:
            spec = find_spec(f"toltec.hooks.{hook}")
            if spec:
                module = module_from_spec(spec)
                spec.loader.exec_module(module)  # type: ignore
                module.register(self)  # type: ignore
            else:
                raise RuntimeError(
                    f"Hook module 'toltec.hooks.{hook}' couldn’t be loaded"
                )

    def __enter__(self) -> "Builder":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.docker.close()

    @util.hook
    def post_parse(self, recipe: Recipe) -> None:
        """
        Triggered after a recipe has been parsed, before executing it.

        :param recipe: recipe object, may be modified by hook listeners
        """

    @util.hook
    def post_fetch_sources(self, recipe: Recipe, src_dir: str) -> None:
        """
        Triggered after the sources of a recipe have been fetched, before
        running the prepare() script.

        :param recipe: recipe object
        :param src_dir: folder in which sources have been extracted
        """

    @util.hook
    def post_prepare(self, recipe: Recipe, src_dir: str) -> None:
        """
        Triggered after the prepare() script of a recipe has been run.

        :param recipe: recipe object
        :param src_dir: folder in which sources have been prepared
        """

    @util.hook
    def post_build(self, recipe: Recipe, src_dir: str) -> None:
        """
        Triggered after a recipe’s artifacts have been built.

        :param recipe: recipe object
        :param src_dir: folder in which artifacts have been built
        """

    @util.hook
    def post_package(
        self, package: Package, src_dir: str, pkg_dir: str
    ) -> None:
        """
        Triggered after part of the artifacts from a build have been copied
        in place to the packaging directory.

        :param package: package object
        :param src_dir: folder in which artifacts have been built
        :param pkg_dir: folder in which artifacts to package have been copied
        """

    @util.hook
    def post_archive(self, package: Package, ar_path: str) -> None:
        """
        Triggered after a package archive has been generated.

        :param package: package object
        :param archive: path to the final package archive
        """

    def make(
        self,
        recipe_bundle: RecipeBundle,
        build_matrix: Optional[Mapping[str, Optional[List[Package]]]] = None,
        check_directory: bool = True,
    ) -> bool:
        """
        Build packages defined by a recipe.

        :param recipe_bundle: architecture versions of the recipe to make
        :param build_matrix: set of packages to build for each architecture
            (default: all supported packages for each architecture)
        :returns: true if all the requested packages were built correctly
        """
        if check_directory and not util.check_directory(
            self.work_dir,
            f"The build directory '{self.work_dir}' \
already exists.\nWould you like to [c]ancel, [r]emove that directory, \
or [k]eep it (not recommended)?",
        ):
            return False

        os.makedirs(self.work_dir, exist_ok=True)
        os.makedirs(self.dist_dir, exist_ok=True)

        for name in (
            list(build_matrix.keys())
            if build_matrix is not None
            else list(recipe_bundle.keys())
        ):
            if not self._make_arch(
                recipe_bundle[name],
                os.path.join(self.work_dir, name),
                build_matrix[name] if build_matrix is not None else None,
            ):
                return False

        return True

    def _make_arch(
        self,
        recipe: Recipe,
        build_dir: str,
        packages: Optional[List[Package]] = None,
    ) -> bool:
        self.post_parse(recipe)

        src_dir = os.path.join(build_dir, "src")
        os.makedirs(src_dir, exist_ok=True)

        self._fetch_sources(recipe, src_dir)
        self.post_fetch_sources(recipe, src_dir)

        self._prepare(recipe, src_dir)
        self.post_prepare(recipe, src_dir)

        self._build(recipe, src_dir)
        self.post_build(recipe, src_dir)

        base_pkg_dir = os.path.join(build_dir, "pkg")
        os.makedirs(base_pkg_dir, exist_ok=True)

        for package in (
            packages if packages is not None else recipe.packages.values()
        ):
            pkg_dir = os.path.join(base_pkg_dir, package.name)
            os.makedirs(pkg_dir, exist_ok=True)

            self._package(package, src_dir, pkg_dir)
            self.post_package(package, src_dir, pkg_dir)

            ar_path = os.path.join(self.dist_dir, package.filename())
            ar_dir = os.path.dirname(ar_path)
            os.makedirs(ar_dir, exist_ok=True)

            self._archive(package, pkg_dir, ar_path)
            self.post_archive(package, ar_path)

        return True

    def _fetch_sources(
        self,
        recipe: Recipe,
        src_dir: str,
    ) -> None:
        """Fetch and extract all source files required to build a recipe."""
        logger.info("Fetching source files")

        for source in recipe.sources:
            filename = os.path.basename(source.url)
            local_path = os.path.join(src_dir, filename)

            if self.URL_REGEX.match(source.url) is None:
                # Get source file from the recipe’s directory
                shutil.copy2(os.path.join(recipe.path, source.url), local_path)
            else:
                # Fetch source file from the network
                req = requests.get(source.url, timeout=(3.05, 300))

                if req.status_code != 200:
                    raise BuildError(
                        f"Unexpected status code while fetching \
source file '{source.url}', got {req.status_code}"
                    )

                with open(local_path, "wb") as local:
                    for chunk in req.iter_content(chunk_size=1024):
                        local.write(chunk)

            # Verify checksum
            file_sha = util.file_sha256(local_path)
            if source.checksum not in ("SKIP", file_sha):
                raise BuildError(
                    f"Invalid checksum for source file {source.url}:\n"
                    f"  expected {source.checksum}\n"
                    f"  actual   {file_sha}"
                )

            # Automatically extract source archives
            if not source.noextract:
                if not util.auto_extract(local_path, src_dir):
                    logger.debug(
                        "Not extracting %s (unsupported archive type)",
                        local_path,
                    )

    @staticmethod
    def _prepare(recipe: Recipe, src_dir: str) -> None:
        """Prepare source files before building."""
        if not recipe.prepare:
            logger.debug("Skipping prepare (nothing to do)")
            return

        logger.info("Preparing source files")
        logs = bash.run_script(
            script=recipe.prepare,
            variables={
                "srcdir": src_dir,
            },
        )
        bash.pipe_logs(logger, logs, "prepare()")

    def _build(self, recipe: Recipe, src_dir: str) -> None:
        """Build artifacts for a recipe."""
        if not recipe.build:
            logger.debug("Skipping build (nothing to do)")
            return

        logger.info("Building artifacts")

        # Set fixed atime and mtime for all the source files
        epoch = int(recipe.timestamp.timestamp())

        for filename in util.list_tree(src_dir):
            os.utime(filename, (epoch, epoch))

        mount_src = "/src"
        repo_src = "/repo"
        uid = os.getuid()
        pre_script: List[str] = []

        # Install required dependencies
        build_deps = []
        host_deps = []

        for dep in recipe.makedepends:
            if dep.kind == DependencyKind.BUILD:
                build_deps.append(dep.package)
            elif dep.kind == DependencyKind.HOST:
                host_deps.append(dep.package)

        if build_deps:
            pre_script.extend(
                (
                    "export DEBIAN_FRONTEND=noninteractive",
                    "apt-get update -qq",
                    "apt-get install -qq --no-install-recommends"
                    ' -o Dpkg::Options::="--force-confdef"'
                    ' -o Dpkg::Options::="--force-confold"'
                    " -- " + " ".join(build_deps),
                )
            )

        if host_deps:
            opkg_conf_path = "$SYSROOT/etc/opkg/opkg.conf"
            pre_script.extend(
                (
                    'echo -n "dest root /',
                    "arch all 100",
                    "arch armv7-3.2 160",
                    "src/gz entware https://bin.entware.net/armv7sf-k3.2",
                    "arch rmall 200",
                    "src/gz toltec-rmall file:///repo/rmall",
                    f'" > "{opkg_conf_path}"',
                )
            )

            if recipe.arch != "rmall":
                pre_script.extend(
                    (
                        f'echo -n "arch {recipe.arch} 250',
                        f"src/gz toltec-{recipe.arch} file:///repo/{recipe.arch}",
                        f'" >> "{opkg_conf_path}"',
                    )
                )

            pre_script.extend(
                (
                    "opkg update --verbosity=0",
                    "opkg install --verbosity=0 --no-install-recommends"
                    " -- " + " ".join(host_deps),
                )
            )

        logs = bash.run_script_in_container(
            self.docker,
            image=self.IMAGE_PREFIX + recipe.image,
            mounts=[
                docker.types.Mount(
                    type="bind",
                    source=os.path.abspath(src_dir),
                    target=mount_src,
                ),
                docker.types.Mount(
                    type="bind",
                    source=os.path.abspath(self.dist_dir),
                    target=repo_src,
                ),
            ],
            variables={
                "srcdir": mount_src,
            },
            script="\n".join(
                (
                    *pre_script,
                    f'cd "{mount_src}"',
                    recipe.build,
                    f'chown -R {uid}:{uid} "{mount_src}"',
                )
            ),
        )
        bash.pipe_logs(logger, logs, "build()")

    @staticmethod
    def _package(package: Package, src_dir: str, pkg_dir: str) -> None:
        """Make a package from a recipe’s build artifacts."""
        logger.info("Packaging build artifacts for %s", package.name)
        logs = bash.run_script(
            script=package.package,
            variables={
                "srcdir": src_dir,
                "pkgdir": pkg_dir,
            },
        )

        bash.pipe_logs(logger, logs, "package()")
        logger.debug("Resulting tree:")

        for filename in util.list_tree(pkg_dir):
            logger.debug(
                " - %s",
                os.path.normpath(
                    os.path.join("/", os.path.relpath(filename, pkg_dir))
                ),
            )

    @staticmethod
    def _archive(package: Package, pkg_dir: str, ar_path: str) -> None:
        """Create an archive for a package."""
        logger.info("Creating archive %s", package.filename())

        # Convert install scripts to Debian format
        scripts = {}
        script_header = textwrap.dedent(
            """\
            #!/usr/bin/env bash
            set -euo pipefail
            """
        )

        for name, script, action in (
            ("preinstall", "preinst", "install"),
            ("configure", "postinst", "configure"),
        ):
            function = getattr(package, name)

            if function:
                scripts[script] = "\n".join(
                    (
                        script_header,
                        textwrap.dedent(
                            f"""\
                            if [[ $1 = {action} ]]; then
                                script() {{
                            """
                        ),
                        function,
                        textwrap.dedent(
                            """\
                                }
                                script
                            fi
                            """
                        ),
                    )
                )

        for step in ("pre", "post"):
            if getattr(package, step + "upgrade") or getattr(
                package, step + "remove"
            ):
                script = script_header

                for action in ("upgrade", "remove"):
                    function = getattr(package, step + action)

                    if function:
                        script += "\n".join(
                            (
                                textwrap.dedent(
                                    f"""\
                                    if [[ $1 = {action} ]]; then
                                        script() {{
                                    """
                                ),
                                function,
                                textwrap.dedent(
                                    """\
                                        }
                                        script
                                    fi
                                    """
                                ),
                            )
                        )

                scripts[step + "rm"] = script

        logger.debug("Install scripts:")

        if scripts:
            for script in sorted(scripts):
                logger.debug(" - %s", script)
        else:
            logger.debug("(none)")

        epoch = int(package.parent.timestamp.timestamp())

        with open(ar_path, "wb") as file:
            ipk.write(
                file,
                epoch=epoch,
                pkg_dir=pkg_dir,
                metadata=package.control_fields(),
                scripts=scripts,
            )

        # Set fixed atime and mtime for the resulting archive
        os.utime(ar_path, (epoch, epoch))



"""
Build hook to automatically add install-lib helper functions

After the artifacts are packaged, this hook will look for known install-lib
methods and add them to scripts if found.
"""
import logging

from typing import Set, Iterable
from toltec.builder import Builder
from toltec.recipe import Package
from toltec.util import listener

logger = logging.getLogger(__name__)

METHODS = {}


def add_method(name: str, src: str, *depends: Iterable[str]) -> None:
    """Add a method to be automatically added to scripts that use it"""
    METHODS[name] = (
        src,
        depends,
    )


def register(builder: Builder) -> None:
    """Register the hook"""

    @listener(builder.post_package)
    def post_package(
        builder: Builder,  # pylint: disable=unused-argument
        package: Package,
        src_dir: str,  # pylint: disable=unused-argument
        pkg_dir: str,  # pylint: disable=unused-argument
    ) -> None:
        for name in (
            "configure",
            "postremove",
            "postupgrade",
            "preinstall",
            "preremove",
            "preupgrade",
        ):
            function = getattr(package, name)
            methods: Set[str] = set()
            for method in METHODS:  # pylint: disable=consider-using-dict-items
                if method in function:
                    methods.add(method)
                    src, depends = METHODS[method]
                    methods.update(depends)  # type: ignore

            for method in methods:
                src, depends = METHODS[method]
                function = f"""
{method}() {{
    {src}
}}
{function}
"""

            setattr(package, name, function)

    add_method("is-enabled", 'systemctl --quiet is-enabled "$1" 2> /dev/null')
    add_method(
        "is-masked",
        '[[ "$(systemctl is-enabled "$1" 2> /dev/null)" == "masked" ]]',
    )
    add_method("is-active", 'systemctl --quiet is-active "$1" 2> /dev/null')
    add_method(
        "get-conflicts",
        """
    # Find enabled units that have a conflicting name
    for name in $(systemctl cat "$1" | awk -F'=' '/^Alias=/{print $2}'); do
        local realname
        if realname="$(basename "$(readlink "/etc/systemd/system/$name")")"; then
            echo "$realname"
        fi
    done

    # Find units that are declared as conflicting
    # (systemd automatically adds a conflict with "shutdown.target" to all
    # service units see systemd.service(5), section "Automatic Dependencies")
    systemctl show "$1" | awk -F'=' '/^Conflicts=/{print $2}' \
        | sed 's|\bshutdown.target\b||'
    """,
    )
    add_method(
        "how-to-enable",
        """
    for conflict in $(get-conflicts "$1"); do
        if is-enabled "$conflict"; then
            echo "$ systemctl disable --now ${conflict/.service/}"
        fi
    done

    echo "$ systemctl enable --now ${1/.service/}"
    """,
        "is-enabled",
        "get-conflicts",
    )
    add_method(
        "reload-oxide-apps",
        """
    if ! is-active tarnish.service; then
        return
    fi
    echo -n "Reloading Oxide applications: "
    local ret
    if type update-desktop-database &> /dev/null; then
        update-desktop-database --quiet
        ret=$?
    else
        /opt/bin/rot apps call reload 2> /dev/null
        ret=$?
    fi
    if [ $ret -eq 0 ]; then
        echo "Done!"
    else
        echo "Failed!"
    fi
    """,
        "is-active",
    )
    add_method(
        "add-bind-mount",
        """
    local unit_name
    local unit_path
    unit_name="$(systemd-escape --path "$2").mount"
    unit_path="/lib/systemd/system/$unit_name"

    if [[ -e $unit_path ]]; then
        echo "Bind mount configuration for '$2' already exists, updating"
    else
        echo "Mounting '$1' over '$2'"
    fi

    cat > "$unit_path" << UNIT
    [Unit]
    Description=Bind mount $1 over $2
    DefaultDependencies=no
    Conflicts=umount.target
    Before=local-fs.target umount.target

    [Mount]
    What=$1
    Where=$2
    Type=none
    Options=bind

    [Install]
    WantedBy=local-fs.target
    UNIT

    systemctl daemon-reload
    systemctl enable "$unit_name"
    systemctl restart "$unit_name"
    """,
    )
    add_method(
        "remove-bind-mount",
        """
    local unit_name
    local unit_path
    unit_name="$(systemd-escape --path "$1").mount"
    unit_path="/lib/systemd/system/$unit_name"

    if [[ ! -e $unit_path ]]; then
        echo "No existing bind mount for '$1'"
        return 1
    fi

    echo "Removing mount over '$1'"
    systemctl disable "$unit_name"
    systemctl stop "$unit_name"
    rm "$unit_path"
    systemctl daemon-reload
    """,
    )
    add_method(
        "unit-exists",
        '[ "$(systemctl --quiet list-unit-files "${1}" | grep -c "${1}")" -eq 1 ]',
    )
    add_method(
        "disable-unit",
        """
    if ! unit-exists "${1}"; then
        return
    fi
    if is-active "$1"; then
        echo "Stopping ${1}"
        systemctl stop "${1}"
    fi
    if is-enabled "${1}"; then
        echo "Disabling ${1}"
        systemctl disable "${1}"
    fi
    """,
        "unit-exists",
        "is-active",
        "is-enabled",
    )



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Recipe parsers."""

from . import bash
from ..recipe import RecipeBundle


parsers = (bash,)


def parse(path: str) -> RecipeBundle:
    """
    Load a recipe defined using any of the supported formats.

    :param path: path to the directory containing the recipe definition
    :raises RecipeError: if there is an error in the recipe
    :raises FileNotFoundError: if no recipe is found at the given path
    :returns: loaded recipe
    """
    for parser in parsers:
        try:
            # See <https://github.com/python/mypy/issues/9841>
            # and <https://github.com/python/mypy/issues/5018>
            return parser.parse(path)  # type: ignore
        except FileNotFoundError:
            pass

    raise FileNotFoundError(
        f"No recipe in a compatible format found in '{path}'"
    )



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Build packages from the recipe in [DIR]."""

import argparse
import os
import sys
from importlib.util import find_spec, spec_from_file_location, module_from_spec
from typing import Dict, List, Optional
from toltec import parse_recipe
from toltec.builder import Builder
from toltec.recipe import Package
from toltec.repo import make_index
from toltec import util


def main() -> int:  # pylint:disable=too-many-branches
    """Execute requested commands and return appropriate exit code."""
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        "recipe_dir",
        metavar="DIR",
        nargs="?",
        default=os.getcwd(),
        help="""path to a directory containing the recipe to build
        (default: current directory)""",
    )

    parser.add_argument(
        "-w",
        "--work-dir",
        metavar="DIR",
        default=os.path.join(os.getcwd(), "build"),
        help="""path to a directory used for building the package
        (default: [current directory]/build)""",
    )

    parser.add_argument(
        "-d",
        "--dist-dir",
        metavar="DIR",
        default=os.path.join(os.getcwd(), "dist"),
        help="""path to a directory where built packages are stored
        (default: [current directory]/dist)""",
    )

    parser.add_argument(
        "-a",
        "--arch-name",
        metavar="ARCHNAME",
        action="append",
        help="""only build for the given architecture (can
        be repeated)""",
    )

    parser.add_argument(
        "-p",
        "--package-name",
        metavar="PACKAGENAME",
        action="append",
        help="""list of packages to build (default: all packages from the
        recipe, can be repeated)""",
    )

    parser.add_argument(
        "-H",
        "--hook",
        metavar="PATH",
        action="append",
        help="""name or path to a Python module that registers listeners for
        build hooks (can be repeated) - if a path is passed, it must start with
        either a dot or a slash""",
    )

    util.argparse_add_verbose(parser)
    util.argparse_add_warning(parser)
    args = parser.parse_args()
    util.setup_logging(args)

    recipe_bundle = parse_recipe(args.recipe_dir)

    with Builder(args.work_dir, args.dist_dir) as builder:
        if args.hook:
            for ident in args.hook:
                if ident and ident[0] in (".", "/"):
                    spec = spec_from_file_location("toltec.hooks.user", ident)
                else:
                    spec = find_spec(ident)

                if spec:
                    module = module_from_spec(spec)
                    spec.loader.exec_module(module)  # type: ignore
                    module.register(builder)  # type: ignore
                else:
                    raise RuntimeError(
                        f"Hook module '{ident}' couldn’t be loaded"
                    )

        build_matrix: Optional[Dict[str, Optional[List[Package]]]] = None

        if args.arch_name or args.package_name:
            build_matrix = {}

            for arch, recipes in recipe_bundle.items():
                if args.package_name:
                    build_matrix[arch] = [
                        recipes.packages[pkg_name]
                        for pkg_name in args.package_name
                    ]
                else:
                    build_matrix[arch] = None

        if not builder.make(recipe_bundle, build_matrix):
            return 1

    make_index(args.dist_dir)
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Build hook for patching all binary objects that access /dev/fb0 to depend on
librm2fb_client.so.1 after building a recipe.

After the build() script is run, and before the artifacts are packaged, this
hook looks for ARM ELF-files in the build directory that access /dev/fb0.
It then uses patchelf to add a dependency on librm2fb_client.so.1 to the
binaries. This behavior is only enabled if the recipe declares the
'patch_rm2fb' flag.
"""
import os
import logging
import shlex
from elftools.elf.elffile import ELFFile
from toltec.builder import Builder
from toltec.recipe import Recipe
from toltec.util import listener
from toltec.hooks.strip import walk_elfs, run_in_container, MOUNT_SRC

logger = logging.getLogger(__name__)


def register(builder: Builder) -> None:
    """Register the hook"""

    @listener(builder.post_build)
    def post_build(  # pylint: disable=too-many-locals
        builder: Builder, recipe: Recipe, src_dir: str
    ) -> None:
        if "patch_rm2fb" not in recipe.flags:
            return

        logger.debug("Adding dependency to rm2fb ('patch_rm2fb' flag is set)")
        # Search for binary objects that can be stripped
        binaries = []

        def filter_elfs(info: ELFFile, file_path: str) -> None:
            symtab = info.get_section_by_name(".symtab")
            if symtab is None or info.get_machine_arch() != "ARM":
                return

            dynamic = info.get_section_by_name(".dynamic")
            rodata = info.get_section_by_name(".rodata")
            if dynamic and rodata and rodata.data().find(b"/dev/fb0") != -1:
                binaries.append(file_path)

        walk_elfs(src_dir, filter_elfs)

        if not binaries:
            logger.debug("Skipping, no arm binaries found")
            return

        # Save original mtimes to restore them afterwards
        # This will prevent any Makefile rules to be triggered again
        # in packaging scripts that use `make install`
        original_mtime = {}

        script = [
            "export DEBIAN_FRONTEND=noninteractive",
            "apt-get update -qq",
            "apt-get install -qq --no-install-recommends patchelf",
        ]

        def docker_file_path(file_path: str) -> str:
            return shlex.quote(
                os.path.join(MOUNT_SRC, os.path.relpath(file_path, src_dir))
            )

        for file_path in binaries:
            original_mtime[file_path] = os.stat(file_path).st_mtime_ns

        script.append(
            "patchelf --add-needed librm2fb_client.so.1 "
            + " ".join(docker_file_path(file_path) for file_path in binaries)
        )

        run_in_container(builder, src_dir, logger, script)

        # Restore original mtimes
        for file_path, mtime in original_mtime.items():
            os.utime(file_path, ns=(mtime, mtime))



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Read and write ipk packages."""

from gzip import GzipFile
from typing import Dict, IO, Optional, Type, Union
from types import TracebackType
from io import BytesIO
import tarfile
import operator
import os


def _targz_open(fileobj: IO[bytes], epoch: int) -> tarfile.TarFile:
    """Open a gzip compressed tar archive for writing."""
    # HACK: Modified code from `tarfile.TarFile.gzopen` to support
    # setting the `mtime` attribute on `GzipFile`
    gzipobj = GzipFile(
        filename="", mode="wb", compresslevel=9, fileobj=fileobj, mtime=epoch
    )

    try:
        # pylint:disable=consider-using-with
        archive = tarfile.TarFile(
            mode="w",
            fileobj=gzipobj,  # type:ignore
            format=tarfile.GNU_FORMAT,
        )
    except:
        gzipobj.close()
        raise

    archive._extfileobj = False  # type:ignore # pylint:disable=protected-access
    return archive


def _clean_info(
    root: Optional[str], epoch: int, info: tarfile.TarInfo
) -> tarfile.TarInfo:
    """
    Remove variable data from an archive entry.

    :param root: absolute path to the root directory from which the
        entry was added, or None to disable turning the name into a
        relative path
    :param epoch: fixed modification time to set
    :param info: tarinfo object to set
    :returns: changed tarinfo
    """
    if root is not None:
        info.name = os.path.relpath("/" + info.name, root)

    if not info.name.startswith("."):
        info.name = "./" + info.name

    info.uid = 0
    info.gid = 0
    info.uname = ""
    info.gname = ""
    info.mtime = epoch

    return info


def _add_file(
    archive: tarfile.TarFile, name: str, mode: int, epoch: int, data: bytes
) -> None:
    """
    Add an in-memory file into a tar archive.

    :param archive: archive to append to
    :param name: name of the file to add
    :param mode: permissions of the file
    :param epoch: fixed modification time to set
    :param data: file contents
    """
    info = tarfile.TarInfo("./" + name)
    info.size = len(data)
    info.mode = mode
    archive.addfile(_clean_info(None, epoch, info), BytesIO(data))


def write_control(
    file: IO[bytes], epoch: int, metadata: str, scripts: Dict[str, str]
) -> None:
    """
    Create the control sub-archive of an ipk package.

    See <https://www.debian.org/doc/debian-policy/ch-controlfields.html>
    and <https://www.debian.org/doc/debian-policy/ch-maintainerscripts.html>.

    :param file: file to which the sub-archive will be written
    :param epoch: fixed modification time to set in the archive metadata
    :param metadata: package metadata (main control file)
    :param scripts: optional maintainer scripts
    """
    with _targz_open(file, epoch) as archive:
        root_info = tarfile.TarInfo("./")
        root_info.type = tarfile.DIRTYPE
        root_info.mode = 0o755
        archive.addfile(_clean_info(None, epoch, root_info))

        _add_file(archive, "control", 0o644, epoch, metadata.encode())

        for name, script in sorted(scripts.items(), key=operator.itemgetter(0)):
            _add_file(archive, name, 0o755, epoch, script.encode())


def write_data(
    file: IO[bytes],
    epoch: int,
    pkg_dir: Optional[str] = None,
) -> None:
    """
    Create the data sub-archive of an ipk package.

    :param file: file to which the sub-archive will be written
    :param epoch: fixed modification time to set in the archive metadata
    :param pkg_dir: directory containing the package tree to include in the
        data sub-archive, leave empty to generate an empty data archive
    """
    with _targz_open(file, epoch) as archive:
        if pkg_dir is not None:
            archive.add(
                pkg_dir, filter=lambda info: _clean_info(pkg_dir, epoch, info)
            )


def write(
    file: IO[bytes],
    epoch: int,
    metadata: str,
    scripts: Dict[str, str],
    pkg_dir: Optional[str] = None,
) -> None:
    """
    Create an ipk package.

    :param file: file to which the package will be written
    :param epoch: fixed modification time to set in the archives metadata
    :param metadata: package metadata (main control file)
    :param scripts: optional maintainer scripts
    :param pkg_dir: directory containing the package tree to include in the
        data sub-archive, leave empty to generate an empty data archive
    """
    with BytesIO() as control, BytesIO() as data, _targz_open(
        file, epoch
    ) as archive:
        root_info = tarfile.TarInfo("./")
        root_info.type = tarfile.DIRTYPE
        archive.addfile(_clean_info(None, epoch, root_info))

        write_control(control, epoch, metadata, scripts)
        _add_file(archive, "control.tar.gz", 0o644, epoch, control.getvalue())

        write_data(data, epoch, pkg_dir)
        _add_file(archive, "data.tar.gz", 0o644, epoch, data.getvalue())

        _add_file(archive, "debian-binary", 0o644, epoch, b"2.0\n")


class Reader:
    """Read from ipk packages."""

    def __init__(self, file: Union[str, IO[bytes]]):
        """
        Create a package reader.
        :param file: path to the package file to read, or opened
            file object for a package file (in the second case, the
            package file object will not by closed on exit)
        """
        self._file: Optional[IO[bytes]] = None

        if isinstance(file, str):
            self._file = open(file, "rb")  # pylint:disable=consider-using-with
            self._close = True
        else:
            self._file = file
            self._close = False

        self._root_archive: Optional[tarfile.TarFile] = None
        self._data_file: Optional[IO[bytes]] = None

        self.data: Optional[tarfile.TarFile] = None
        self.metadata: Optional[str] = None
        self.scripts: Dict[str, str] = {}

    def __enter__(self) -> "Reader":
        """Load package data to memory."""
        root_archive = tarfile.TarFile.open(fileobj=self._file)
        control_file = root_archive.extractfile("./control.tar.gz")
        assert control_file is not None

        with control_file:
            with tarfile.TarFile.open(fileobj=control_file) as control_archive:
                for member in control_archive.getmembers():
                    if member.isfile():
                        file = control_archive.extractfile(member)
                        assert file is not None
                        with file:
                            contents = file.read().decode("utf-8")
                            if member.name == "./control":
                                self.metadata = contents
                            else:
                                self.scripts[member.name[2:]] = contents

        data_file = root_archive.extractfile("./data.tar.gz")
        assert data_file is not None
        self._root_archive = root_archive
        self._data_file = data_file
        self.data = tarfile.TarFile.open(fileobj=data_file)

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_inst: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Free resources containing package data."""
        if self.data is not None:
            self.data.close()
            self.data = None

        if self._data_file is not None:
            self._data_file.close()
            self._data_file = None

        if self._root_archive is not None:
            self._root_archive.close()
            self._root_archive = None

        if self._file is not None:
            if self._close:
                self._file.close()

            self._file = None



"""
Built in hooks
"""
__all__ = [
    "patch_rm2fb",
    "strip",
    "reload_oxide_apps",  # Depends on install_lib
    # This hook needs to come after any hooks that may depend on it
    "install_lib",
]



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""Bridge Bash with Python."""

import os
import shlex
import subprocess
import logging
from collections import deque
from typing import Deque, Dict, Generator, List, Optional, Tuple, Union
from docker.client import DockerClient

AssociativeArray = Dict[str, str]
IndexedArray = List[Optional[str]]
LogGenerator = Generator[str, None, None]
Any = Union[str, AssociativeArray, IndexedArray]
Variables = Dict[str, Optional[Any]]
Functions = Dict[str, str]


class ScriptError(Exception):
    """Raised when a launched Bash script exits with a non-zero code."""


# Variables which are defined by default by Bash.  Those variables are excluded
# from the result of `get_declarations()`. Subset of the list at:
# <https://www.gnu.org/software/bash/manual/html_node/Bash-Variables.html>
default_variables = {
    "_",
    "BASH",
    "BASH_ALIASES",
    "BASH_ARGC",
    "BASH_ARGV",
    "BASH_ARGV0",
    "BASH_CMDS",
    "BASH_COMMAND",
    "BASH_LINENO",
    "BASH_LOADABLES_PATH",
    "BASH_SOURCE",
    "BASH_SUBSHELL",
    "BASH_VERSINFO",
    "BASH_VERSION",
    "BASHOPTS",
    "BASHPID",
    "COLUMNS",
    "COMP_WORDBREAKS",
    "DIRSTACK",
    "EPOCHREALTIME",
    "EPOCHSECONDS",
    "EUID",
    "FUNCNAME",
    "GROUPS",
    "HISTCMD",
    "HISTFILE",
    "HISTFILESIZE",
    "HISTSIZE",
    "HOSTNAME",
    "HOSTTYPE",
    "IFS",
    "LINENO",
    "LINES",
    "MACHTYPE",
    "MAILCHECK",
    "OLDPWD",
    "OPTERR",
    "OPTIND",
    "OSTYPE",
    "PATH",
    "PIPESTATUS",
    "PPID",
    "PS1",
    "PS2",
    "PS4",
    "PWD",
    "RANDOM",
    "SECONDS",
    "SHELL",
    "SHELLOPTS",
    "SHLVL",
    "SRANDOM",
    "TERM",
    "UID",
}


def _get_bash_stdout(src: str) -> str:
    """
    Get the stdout from a bash script

    :param src: bash script to run
    :returns: the stdout of the script
    """
    env: Dict[str, str] = {
        "PATH": os.environ["PATH"],
    }

    subshell = subprocess.run(  # pylint:disable=subprocess-run-check
        ["/usr/bin/env", "bash"],
        input=src.encode(),
        capture_output=True,
        env=env,
    )

    errors = subshell.stderr.decode()

    if subshell.returncode == 2 or "syntax error" in errors:
        raise ScriptError(f"Bash syntax error\n{errors}")

    if subshell.returncode != 0 or errors:
        raise ScriptError(f"Bash error\n{errors}")

    return subshell.stdout.decode()


def get_declarations(src: str) -> Tuple[Variables, Functions]:
    """
    Extract all variables and functions defined by a Bash script.

    If a function or a variable is defined or assigned multiple times
    in the script, only the final value is extracted. The script must not
    output anything on the standard output stream.

    :param src: source string of the considered Bash string
    :returns: a tuple containing the declared variables and functions
    """
    src += """
declare -f
declare -p
"""
    declarations = _get_bash_stdout(src)

    # Parse `declare` statements and function statements
    lexer = shlex.shlex(declarations, posix=True)
    lexer.wordchars = lexer.wordchars + "-"

    variables: Variables = {}
    functions: Functions = {}

    while True:
        token = lexer.get_token()

        if token == lexer.eof:
            break

        next_token = lexer.get_token()
        assert token is not None
        assert next_token is not None

        if token == "declare" and next_token[0] == "-":
            lexer.push_token(next_token)
            name, value = _parse_var(lexer)

            if name not in default_variables:
                variables[name] = value
        else:
            if next_token != "(":
                raise ScriptError(
                    f"Unexpected token '{next_token}' on line {lexer.lineno}. Expecting '('."
                )

            _token = lexer.get_token()
            if _token != ")":
                raise ScriptError(
                    f"Unexpected token '{_token}' on line {lexer.lineno}. Expecting ')'."
                )

            start, end = _parse_func(lexer)
            functions[token] = declarations[start:end].strip(" ")

    return variables, functions


def put_variables(variables: Variables) -> str:
    """
    Generate a Bash script fragment which defines a set of variables.

    :param variables: set of variables to define
    :returns: generated Bash fragment
    """
    result = ""

    for name, value in variables.items():
        if value is None:
            result += f"declare -- {name}\n"
        elif isinstance(value, str):
            result += f"declare -- {name}={_generate_string(value)}\n"
        elif isinstance(value, list):
            result += f"declare -a {name}={_generate_indexed(value)}\n"
        elif isinstance(value, dict):
            result += f"declare -A {name}={_generate_assoc(value)}\n"
        else:
            raise ValueError(
                f"Unsupported type {type(value)} for variable \
{name}"
            )

    return result


def put_functions(functions: Functions) -> str:
    """
    Generate a Bash script which defines a set of functions.

    :param functions: set of functions to define
    :returns: generated Bash fragment
    """
    result = ""

    for name, value in functions.items():
        result += f"{name}() {{\n{value}\n}}\n"

    return result


def _parse_string(token: str) -> str:
    """Remove escape sequences from a Bash string."""
    return token.replace("\\$", "$")


def _generate_string(string: str) -> str:
    """Generate a Bash string."""
    return shlex.quote(string)


def _parse_indexed(lexer: shlex.shlex) -> IndexedArray:
    """Parse an indexed Bash array."""
    assert lexer.get_token() == "("
    result: List[Optional[str]] = []

    while True:
        token = lexer.get_token()
        assert token != lexer.eof

        if token == ")":
            break

        assert token == "["
        index = int(lexer.get_token() or "")
        assert lexer.get_token() == "]"
        assert lexer.get_token() == "="
        string_token = lexer.get_token() or ""
        if string_token == "$":
            string_token = lexer.get_token() or ""
        value = _parse_string(string_token)

        # Grow the result array so that the index exists
        if index >= len(result):
            result.extend([None] * (index - len(result) + 1))

        result[index] = value

    return result


def _generate_indexed(array: IndexedArray) -> str:
    """Generate an indexed Bash array."""
    return (
        "("
        + " ".join(
            f"[{index}]={_generate_string(value)}"
            for index, value in enumerate(array)
            if value is not None
        )
        + ")"
    )


def _parse_assoc(lexer: shlex.shlex) -> AssociativeArray:
    """Parse an associative Bash array."""
    assert lexer.get_token() == "("
    result: AssociativeArray = {}

    while True:
        token = lexer.get_token()
        assert token != lexer.eof

        if token == ")":
            break

        assert token == "["
        key = lexer.get_token()
        assert key is not None
        assert lexer.get_token() == "]"
        assert lexer.get_token() == "="
        string_token = lexer.get_token() or ""
        if string_token == "$":
            string_token = lexer.get_token() or ""
        value = _parse_string(string_token)

        result[key] = value

    return result


def _generate_assoc(array: AssociativeArray) -> str:
    """Generate an associative Bash array."""
    return (
        "("
        + " ".join(
            f"[{_generate_string(key)}]={_generate_string(value)}"
            for key, value in array.items()
        )
        + ")"
    )


def _parse_var(lexer: shlex.shlex) -> Tuple[str, Optional[Any]]:
    """Parse a variable declaration."""
    flags_token = lexer.get_token()
    assert flags_token is not None

    if flags_token != "--":
        var_flags = set(flags_token[1:])
    else:
        var_flags = set()

    var_name = lexer.get_token()
    var_value: Optional[Any] = None
    lookahead = lexer.get_token()
    assert var_name is not None
    assert lookahead is not None

    if lookahead == "=":
        if "a" in var_flags:
            var_value = _parse_indexed(lexer)
        elif "A" in var_flags:
            var_value = _parse_assoc(lexer)
        else:
            string_token = lexer.get_token() or ""
            if string_token == "$":
                quoted_string = lexer.get_token() or ""
                string_token = _get_bash_stdout(
                    "echo -n $" + shlex.quote(quoted_string)
                )
            var_value = _parse_string(string_token)
    else:
        lexer.push_token(lookahead)

    return var_name, var_value


def _parse_func(lexer: shlex.shlex) -> Tuple[int, int]:
    """Find the starting and end bounds of a function declaration."""
    assert lexer.get_token() == "{"
    brace_depth = 1

    start_byte = lexer.instream.tell()

    while brace_depth > 0:
        token = lexer.get_token()
        assert token != lexer.eof

        if token == "{":
            brace_depth += 1
        elif token == "}":
            brace_depth -= 1

    end_byte = lexer.instream.tell() - 1
    return start_byte, end_byte


def run_script(variables: Variables, script: str) -> LogGenerator:
    """
    Run a Bash script and stream its output.

    :param variables: Bash variables to set before running the script
    :param script: Bash script to execute
    :returns: generator yielding output lines from the script
    :raises ScriptError: if the script exits with a non-zero code
    """
    with subprocess.Popen(
        ["/usr/bin/env", "bash"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    ) as process:
        assert process.stdin is not None
        assert process.stdout is not None
        process.stdin.write(
            "\n".join(
                (
                    "set -euo pipefail",
                    put_variables(variables),
                    "script() {",
                    script,
                    "}",
                    "script",
                )
            ).encode()
        )
        process.stdin.close()

        while process.poll() is None:
            line = process.stdout.readline()
            if line:
                yield line.decode().strip()

        if process.returncode != 0:
            raise ScriptError(f"Script exited with code {process.returncode}")


def run_script_in_container(
    docker: DockerClient,
    image: str,
    mounts: List,
    variables: Variables,
    script: str,
) -> LogGenerator:
    """
    Run a Bash script inside a Docker container and stream its output.

    :param docker: Docker client
    :param image: image to use for the new container
    :param mounts: paths to mount in the container
    :param variables: Bash variables to set before running the script
    :param script: Bash script to execute
    :returns: generator yielding output lines from the script
    :raises ScriptError: if the script exits with a non-zero code
    """
    container = docker.containers.run(
        image,
        mounts=mounts,
        command=[
            "/usr/bin/env",
            "bash",
            "-c",
            "\n".join(
                (
                    "set -euo pipefail",
                    put_variables(variables),
                    "script() {",
                    script,
                    "}",
                    "script",
                )
            ),
        ],
        detach=True,
        security_opt=["label=disable"],
    )

    try:
        for line in container.logs(stream=True):
            if line:
                yield line.decode().strip()

        result = container.wait()

        if result["StatusCode"] != 0:
            raise ScriptError(f"Script exited with code {result['StatusCode']}")
    finally:
        container.remove()


def pipe_logs(
    logger: logging.Logger,
    logs: LogGenerator,
    prefix: str = "",
    max_lines_on_fail: int = 50,
) -> None:
    """
    Pipe logs from a script to the debug output of a Python logger.

    Print the last :param:`max_lines_on_fail` log lines to the error output in
    case a ScriptError is caught.

    :param logs: generator of log lines
    :param prefix: log prefix
    :param max_lines_on_fail: number of context lines to print
        in non-debug mode
    """
    log_buffer: Deque[str] = deque()

    try:
        for line in logs:
            if logger.getEffectiveLevel() <= logging.DEBUG:
                logger.debug("%s%s", prefix + ": ", line)
            else:
                if len(log_buffer) == max_lines_on_fail:
                    log_buffer.popleft()
                log_buffer.append(line)
    except ScriptError as err:
        if len(log_buffer) > 0:
            logger.info(
                "Only showing up to %s lines of context. "
                "Use --verbose for the full output.",
                max_lines_on_fail,
            )
            for line in log_buffer:
                logger.error("%s%s", prefix + ": ", line)

        if prefix:
            logger.error("%s failed", prefix)

        raise err



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT

from os import path
import unittest
from tempfile import TemporaryDirectory
import subprocess


def filter_index_entry(index_entry):
    return "\n".join(
        item
        for item in index_entry.split("\n")
        if not (item.startswith("SHA256sum:") or item.startswith("Size:"))
    )


class TestBuild(unittest.TestCase):
    def setUp(self) -> None:
        self.dir = path.dirname(path.realpath(__file__))
        self.fixtures_dir = path.join(self.dir, "fixtures")

    def test_build_rmkit(self) -> None:
        with TemporaryDirectory() as tmp_dir:
            rec_dir = path.join(self.fixtures_dir, "rmkit")
            work_dir = path.join(tmp_dir, "build")
            dist_dir = path.join(tmp_dir, "dist")

            result = subprocess.run(
                [
                    "python3",
                    "-m",
                    "toltec",
                    "--work-dir",
                    work_dir,
                    "--dist-dir",
                    dist_dir,
                    "--",
                    rec_dir,
                ],
                capture_output=True,
                check=False,
            )
            self.assertEqual(
                result.returncode, 0, result.stderr.decode("utf-8")
            )
            self.assertEqual(result.stdout.decode("utf-8"), "")
            self.assertEqual(
                result.stderr.decode("utf-8"),
                """\
[    INFO] toltec.builder: Fetching source files
[    INFO] toltec.builder: Preparing source files
[    INFO] toltec.builder: Building artifacts
[    INFO] toltec.builder: Packaging build artifacts for bufshot
[    INFO] toltec.builder: Creating archive rmall/bufshot_0.1.0-5_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for genie
[    INFO] toltec.builder: Creating archive rmall/genie_0.1.5-3_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for harmony
[    INFO] toltec.builder: Creating archive rmall/harmony_0.1.3-3_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for iago
[    INFO] toltec.builder: Creating archive rmall/iago_0.1.0-4_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for lamp
[    INFO] toltec.builder: Creating archive rmall/lamp_0.1.0-4_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for mines
[    INFO] toltec.builder: Creating archive rmall/mines_0.1.2-4_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for nao
[    INFO] toltec.builder: Creating archive rmall/nao_0.1.2-3_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for remux
[    INFO] toltec.builder: Creating archive rmall/remux_0.1.9-4_rmall.ipk
[    INFO] toltec.builder: Packaging build artifacts for simple
[    INFO] toltec.builder: Creating archive rmall/simple_0.1.4-3_rmall.ipk
[    INFO] toltec.repo: Generating package index
""",
            )

            with open(path.join(dist_dir, "rmall", "Packages"), "r") as index:
                index_data = index.read()
                index_entries = set(
                    filter_index_entry(entry)
                    for entry in index_data.split("\n\n")
                    if entry
                )

                self.assertEqual(
                    index_entries,
                    {
                        """Package: simple
Description: Simple app script for writing scripted applications
Homepage: https://rmkit.dev/apps/sas
Version: 0.1.4-3
Section: devel
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: simple_0.1.4-3_rmall.ipk""",
                        """Package: remux
Description: Launcher that supports multi-tasking applications
Homepage: https://rmkit.dev/apps/remux
Version: 0.1.9-4
Section: launchers
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: remux_0.1.9-4_rmall.ipk""",
                        """Package: nao
Description: Nao Package Manager: opkg UI built with SAS
Homepage: https://rmkit.dev/apps/nao
Version: 0.1.2-3
Section: admin
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display, simple
Filename: nao_0.1.2-3_rmall.ipk""",
                        """Package: mines
Description: Mine detection game
Homepage: https://rmkit.dev/apps/minesweeper
Version: 0.1.2-4
Section: games
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: mines_0.1.2-4_rmall.ipk""",
                        """Package: lamp
Description: config based stroke injection utility
Homepage: https://rmkit.dev/apps/lamp
Version: 0.1.0-4
Section: utils
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: lamp_0.1.0-4_rmall.ipk""",
                        """Package: iago
Description: overlay for drawing shapes via stroke injection
Homepage: https://rmkit.dev/apps/iago
Version: 0.1.0-4
Section: utils
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display, lamp
Filename: iago_0.1.0-4_rmall.ipk""",
                        """Package: harmony
Description: Procedural sketching app
Homepage: https://rmkit.dev/apps/harmony
Version: 0.1.3-3
Section: drawing
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: harmony_0.1.3-3_rmall.ipk""",
                        """Package: genie
Description: Gesture engine that connects commands to gestures
Homepage: https://rmkit.dev/apps/genie
Version: 0.1.5-3
Section: utils
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: genie_0.1.5-3_rmall.ipk""",
                        """Package: bufshot
Description: program for saving the framebuffer as a png
Homepage: https://github.com/rmkit-dev/rmkit/tree/master/src/bufshot
Version: 0.1.0-5
Section: utils
Maintainer: raisjn <of.raisjn@gmail.com>
License: MIT
Architecture: rmall
Depends: display
Filename: bufshot_0.1.0-5_rmall.ipk""",
                    },
                )



# Copyright (c) 2021 The Toltec Contributors
# SPDX-License-Identifier: MIT
"""
Package recipes.

A package is a final user-installable software archive. A recipe contains
the metadata and instructions necessary to build one or more related
packages (in the latter case, it is called a split package).
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, NamedTuple, Set
import os
import textwrap
from .version import Version, Dependency


class RecipeError(Exception):
    """Raised when a recipe definition is malformed."""

    def __init__(self, path: str, message: str):
        super().__init__()
        self.path = path
        self.message = message

    def __repr__(self) -> str:
        return f"RecipeError(path={repr(self.path)}, \
message={repr(self.message)})"

    def __str__(self) -> str:
        return f"{self.path}: {self.message}"


class RecipeWarning(Warning):
    """Raised for less serious issues in recipe definitions."""


# Set of variations of the same recipes that target different architectures
RecipeBundle = Dict[str, "Recipe"]


class Source(NamedTuple):
    """Source item needed to build a recipe."""

    # URL or local relative path to the source item
    url: str

    # SHA-256 checksum for validating the source integrity
    checksum: str

    # If true, do not attempt to extract this item after downloading
    noextract: bool


@dataclass
class Recipe:  # pylint: disable=too-many-instance-attributes
    """Recipe declaring how a set of package can be built."""

    # Path to the directory in which the recipe is defined
    path: str

    # Source modification timestamp
    timestamp: datetime

    # Set of source items to be downloaded
    sources: Set[Source]

    # Set of packages that are needed to build this recipe
    makedepends: Set[Dependency]

    # Full name and email address of this recipe’s maintainer
    maintainer: str

    # Docker image used to build this recipe
    image: str

    # Architecture that this recipe targets
    arch: str

    # Set of flags to be used by the build system
    flags: List[str]

    # Bash script for preparing (patching, moving) source files before build
    prepare: str

    # Bash script for building from source
    build: str

    # Set of packages to generate from the build artifacts
    packages: Dict[str, "Package"]


@dataclass
class Package:  # pylint: disable=too-many-instance-attributes
    """Installable package containing build artifacts."""

    # Name of this package, unique among all recipes of a repository
    name: str

    # Recipe used to generate this package
    parent: Recipe

    # Version number
    version: Version

    # Short description
    desc: str

    # URL to the homepage where more information can be found on this package
    url: str

    # Name of the section to which this package belongs
    section: str

    # Identifier for this package’s license
    license: str

    # Set of packages that must be installed for this package to work
    installdepends: Set[Dependency]

    # Set of packages that this package recommends installing
    recommends: Set[Dependency]

    # Set of packages that provide additional features for this package
    optdepends: Set[Dependency]

    # Set of incompatible packages
    conflicts: Set[Dependency]

    # Set of packages replaced by this package
    replaces: Set[Dependency]

    # Set of packages that this package provides
    provides: Set[Dependency]

    # Bash script for packaging build artifacts
    package: str

    # Bash script executed before this package is unpacked
    preinstall: str

    # Bash script executed after this package is unpacked
    configure: str

    # Bash script executed before this package is replaced with a newer version
    preupgrade: str

    # Bash script executed after this package is replaced with a newer version
    postupgrade: str

    # Bash script executed before this package is removed
    preremove: str

    # Bash script executed after this package is removed
    postremove: str

    def pkgid(self) -> str:
        """Get the unique identifier of this package."""
        return "_".join(
            (self.name, str(self.version).replace(":", "_"), self.parent.arch)
        )

    def filename(self) -> str:
        """Get the name of the archive corresponding to this package."""
        return os.path.join(self.parent.arch, self.pkgid() + ".ipk")

    def control_fields(self) -> str:
        """Get the control fields for this package."""
        control = textwrap.dedent(
            f"""\
            Package: {self.name}
            Description: {self.desc}
            Homepage: {self.url}
            Version: {self.version}
            Section: {self.section}
            Maintainer: {self.parent.maintainer}
            License: {self.license}
            Architecture: {self.parent.arch}
            """
        )

        for debian_name, field in (
            ("Depends", self.installdepends),
            ("Recommends", self.recommends),
            ("Suggests", self.optdepends),
            ("Conflicts", self.conflicts),
            ("Replaces", self.replaces),
            ("Provides", self.provides),
        ):
            if field:
                control += (
                    debian_name
                    + ": "
                    + ", ".join(sorted(dep.to_debian() for dep in field if dep))
                    + "\n"
                )

        return control



# Copyright (c) 2022 The Toltec Contributors
# SPDX-License-Identifier: MIT

import unittest
from toltec.version import (
    Version,
    Dependency,
    DependencyKind,
    VersionComparator,
    InvalidVersionError,
    InvalidDependencyError,
)


class TestVersion(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(
            InvalidVersionError,
            msg="Invalid epoch '-1', only non-negative values are allowed",
        ):
            version = Version(-1, "test", "0")

        with self.assertRaises(
            InvalidVersionError,
            msg="Invalid chars in upstream version 't:est', allowed chars "
            "are A-Za-z0-9.+~-",
        ):
            version = Version(0, "t:est", "0")

        with self.assertRaises(
            InvalidVersionError,
            msg="Upstream version cannot be empty",
        ):
            version = Version(0, "", "0")

        with self.assertRaises(
            InvalidVersionError,
            msg="Invalid chars in revision '1-2-3', allowed chars are "
            "A-Za-z0-9.+~",
        ):
            version = Version(0, "test", "1-2-3")

        with self.assertRaises(
            InvalidVersionError,
            msg="Revision cannot be empty",
        ):
            version = Version(0, "test", "")

        version = Version(1, "test", "1")
        self.assertEqual(version.upstream, "test")
        self.assertEqual(version.revision, "1")
        self.assertEqual(version.epoch, 1)

    def test_parse(self):
        valid_versions = (
            ("0.0.0-1", Version(0, "0.0.0", "1")),
            ("0.0.1-1", Version(0, "0.0.1", "1")),
            ("0.1.0-3", Version(0, "0.1.0", "3")),
            ("0.1.1", Version(0, "0.1.1", "0")),
            ("1.0-0", Version(0, "1.0", "0")),
            ("1.0.0", Version(0, "1.0.0", "0")),
            ("1-0-0", Version(0, "1-0", "0")),
            ("1:0.0.14-1", Version(1, "0.0.14", "1")),
            ("1.0.20210219-2", Version(0, "1.0.20210219", "2")),
            ("1.3.5-14", Version(0, "1.3.5", "14")),
            ("19.21-2", Version(0, "19.21", "2")),
            ("2.0.10-1", Version(0, "2.0.10", "1")),
            ("2020.11.08-2", Version(0, "2020.11.08", "2")),
        )

        for string, parsed in valid_versions:
            self.assertEqual(Version.parse(string), parsed)

        with self.assertRaises(
            InvalidVersionError,
            msg="Invalid epoch 'test', must be numeric",
        ):
            version = Version.parse("test:1.1")

        with self.assertRaises(
            InvalidVersionError,
            msg="Upstream version cannot be empty",
        ):
            version = Version.parse("0:-1")

    def test_compare(self):
        self.assertEqual(Version(0, "1.0", "1"), Version(0, "1.0", "1"))

        ordered_pairs = (
            (Version(0, "1.0", "1"), Version(1, "0.1", "1")),
            (Version(0, "1.0", "1"), Version(0, "1.0", "2")),
            (Version(0, "1.0", "2"), Version(0, "1.1", "1")),
            (Version(1, "1.0~~", "7"), Version(1, "1.0~~a", "1")),
            (Version(1, "1.0~~a", "7"), Version(1, "1.0~", "1")),
            (Version(1, "1.0~", "7"), Version(1, "1.0", "1")),
            (Version(1, "1.0", "7"), Version(1, "1.0a", "1")),
        )

        for lower, greater in ordered_pairs:
            self.assertTrue(lower < greater)
            self.assertTrue(lower <= greater)
            self.assertTrue(greater > lower)
            self.assertTrue(greater >= lower)
            self.assertFalse(greater < lower)
            self.assertFalse(greater <= lower)
            self.assertFalse(lower > greater)
            self.assertFalse(lower >= greater)


class TestDependency(unittest.TestCase):
    def test_init(self):
        version = Version(1, "0.1", "1")
        dep = Dependency(
            DependencyKind.BUILD,
            "test",
            VersionComparator.GREATER_THAN_OR_EQUAL,
            version,
        )
        self.assertEqual(dep.kind, DependencyKind.BUILD)
        self.assertEqual(dep.package, "test")
        self.assertEqual(
            dep.version_comparator,
            VersionComparator.GREATER_THAN_OR_EQUAL,
        )
        self.assertEqual(dep.version, version)

    def test_parse(self):
        valid_deps = (
            (
                "test",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.EQUAL,
                    None,
                ),
            ),
            (
                "host:test",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.EQUAL,
                    None,
                ),
            ),
            (
                "build:test",
                Dependency(
                    DependencyKind.BUILD,
                    "test",
                    VersionComparator.EQUAL,
                    None,
                ),
            ),
            (
                "test=0.1-1",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.EQUAL,
                    Version(0, "0.1", "1"),
                ),
            ),
            (
                "test<<0.1-1",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.LOWER_THAN,
                    Version(0, "0.1", "1"),
                ),
            ),
            (
                "test<=0.1-1",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.LOWER_THAN_OR_EQUAL,
                    Version(0, "0.1", "1"),
                ),
            ),
            (
                "test>>0.1-1",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.GREATER_THAN,
                    Version(0, "0.1", "1"),
                ),
            ),
            (
                "test>=0.1-1",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.GREATER_THAN_OR_EQUAL,
                    Version(0, "0.1", "1"),
                ),
            ),
            (
                "test=1:0.1-1",
                Dependency(
                    DependencyKind.HOST,
                    "test",
                    VersionComparator.EQUAL,
                    Version(1, "0.1", "1"),
                ),
            ),
            (
                "build:test=1:0.1-1",
                Dependency(
                    DependencyKind.BUILD,
                    "test",
                    VersionComparator.EQUAL,
                    Version(1, "0.1", "1"),
                ),
            ),
        )

        for string, parsed in valid_deps:
            self.assertEqual(Dependency.parse(string), parsed)

        with self.assertRaises(
            InvalidDependencyError,
            msg="Unknown dependency type 'invalid', valid types are "
            "'build', 'host'",
        ):
            dep = Dependency.parse("invalid:test")

        with self.assertRaises(
            InvalidDependencyError,
            msg="Invalid version comparator '<<>>', valid operators are "
            "'<<', '<=', '=', '>=', '>>'",
        ):
            dep = Dependency.parse("host:test<<>>0.1")

    def test_to_debian(self):
        converted_versions = (
            ("test", "test"),
            ("host:test", "test"),
            ("test=0.1-1", "test (= 0.1-1)"),
            ("test<<0.1-1", "test (<< 0.1-1)"),
            ("test<=0.1-1", "test (<= 0.1-1)"),
            ("test>>0.1-1", "test (>> 0.1-1)"),
            ("test>=0.1-1", "test (>= 0.1-1)"),
            ("test=1:0.1-1", "test (= 1:0.1-1)"),
        )

        for ours, debian in converted_versions:
            self.assertEqual(Dependency.parse(ours).to_debian(), debian)

    def test_match(self):
        matches = (
            ("test", "0.1", True),
            ("test", "0.1-0", True),
            ("test", "0:0.1-0", True),
            ("test", "0.1-1", True),
            ("test", "0.0", True),
            ("test", "1:0.0", True),
            ("test=0.1", "0.1", True),
            ("test=0.1", "0.1-0", True),
            ("test=0.1", "0:0.1-0", True),
            ("test=0.1", "0.1-1", False),
            ("test=0.1", "0.0", False),
            ("test=0.1", "1:0.0", False),
            ("test>=0.1", "0.1", True),
            ("test>=0.1", "0.1-0", True),
            ("test>=0.1", "0:0.1-0", True),
            ("test>=0.1", "0.1-1", True),
            ("test>=0.1", "0.0", False),
            ("test>=0.1", "1:0.0", True),
            ("test>>0.1", "0.1", False),
            ("test>>0.1", "0.1-0", False),
            ("test>>0.1", "0:0.1-0", False),
            ("test>>0.1", "0.1-1", True),
            ("test>>0.1", "0.0", False),
            ("test>>0.1", "1:0.0", True),
            ("test<=0.1", "0.1", True),
            ("test<=0.1", "0.1-0", True),
            ("test<=0.1", "0:0.1-0", True),
            ("test<=0.1", "0.1-1", False),
            ("test<=0.1", "0.0", True),
            ("test<=0.1", "1:0.0", False),
            ("test<<0.1", "0.1", False),
            ("test<<0.1", "0.1-0", False),
            ("test<<0.1", "0:0.1-0", False),
            ("test<<0.1", "0.1-1", False),
            ("test<<0.1", "0.0", True),
            ("test<<0.1", "1:0.0", False),
        )

        for dep, version, result in matches:
            self.assertEqual(
                Dependency.parse(dep).match(Version.parse(version)),
                result,
                f"match {dep} with {version} is {result}",
            )
